@@ -25,12 +25,6 @@ static pool *proto_pool;
 static list protocol_list;
 static list proto_list;
 
-#define WALK_PROTO_LIST(p) do {							\
-	node *nn;								\
-	WALK_LIST(nn, proto_list) {						\
-		struct proto *p = SKIP_BACK(struct proto, glob_node, nn);
-#define WALK_PROTO_LIST_END } } while(0)
-
 #define PD(pr, msg, args...) do { if (pr->debug & D_STATES) { log(L_TRACE "%s: " msg, pr->name , ## args); } } while(0)
 
 list active_proto_list;
@@ -57,7 +51,7 @@ proto_enqueue(list *l, struct proto *p)
 static void
 proto_relink(struct proto *p)
 {
-  list *l;
+  list *l = NULL;
 
   if (p->debug & D_STATES)
     {
@@ -73,6 +67,10 @@ proto_relink(struct proto *p)
   rem_node(&p->n);
   switch (p->core_state)
     {
+    case FS_HUNGRY:
+      l = &inactive_proto_list;
+      break;
+    case FS_FEEDING:
     case FS_HAPPY:
       l = &active_proto_list;
       break;
@@ -80,7 +78,7 @@ proto_relink(struct proto *p)
       l = &flush_proto_list;
       break;
     default:
-      l = &inactive_proto_list;
+      ASSERT(0);
     }
   proto_enqueue(l, p);
 }
@@ -107,6 +105,7 @@ proto_new(struct proto_config *c, unsigned size)
 
   p->cf = c;
   p->debug = c->debug;
+  p->mrtdump = c->mrtdump;
   p->name = c->name;
   p->preference = c->preference;
   p->disabled = c->disabled;
@@ -114,7 +113,6 @@ proto_new(struct proto_config *c, unsigned size)
   p->table = c->table->table;
   p->in_filter = c->in_filter;
   p->out_filter = c->out_filter;
-  p->min_scope = SCOPE_SITE;
   p->hash_key = random_u32();
   c->proto = p;
   return p;
@@ -197,6 +195,7 @@ proto_config_new(struct protocol *pr, unsigned size)
   c->out_filter = FILTER_REJECT;
   c->table = c->global->master_rtc;
   c->debug = new_config->proto_default_debug;
+  c->mrtdump = new_config->proto_default_mrtdump;
   return c;
 }
 
@@ -263,6 +262,71 @@ proto_init(struct proto_config *c)
   return q;
 }
 
+static int
+proto_reconfigure(struct proto *p, struct proto_config *oc, struct proto_config *nc, int type)
+{
+  /* If the protocol is DOWN, we just restart it */
+  if (p->proto_state == PS_DOWN)
+    return 0;
+
+  /* If there is a too big change in core attributes, ... */
+  if ((nc->protocol != oc->protocol) ||
+      (nc->disabled != oc->disabled) ||
+      (nc->table->table != oc->table->table) ||
+      (proto_get_router_id(nc) != proto_get_router_id(oc)))
+    return 0;
+
+  int import_changed = (type != RECONFIG_SOFT) && ! filter_same(nc->in_filter, oc->in_filter);
+  int export_changed = (type != RECONFIG_SOFT) && ! filter_same(nc->out_filter, oc->out_filter);
+
+  /* We treat a change in preferences by reimporting routes */
+  if (nc->preference != oc->preference)
+    import_changed = 1;
+
+  /* If the protocol in not UP, it has no routes and we can ignore such changes */
+  if (p->proto_state != PS_UP)
+    import_changed = export_changed = 0;
+
+  /* Without this hook we cannot reload routes and have to restart the protocol */
+  if (import_changed && ! p->reload_routes)
+    return 0;
+
+  p->debug = nc->debug;
+  p->mrtdump = nc->mrtdump;
+
+  /* Execute protocol specific reconfigure hook */
+  if (! (p->proto->reconfigure && p->proto->reconfigure(p, nc)))
+    return 0;
+
+  DBG("\t%s: same\n", oc->name);
+  PD(p, "Reconfigured");
+  p->cf = nc;
+  p->name = nc->name;
+  p->in_filter = nc->in_filter;
+  p->out_filter = nc->out_filter;
+  p->preference = nc->preference;
+
+  if (import_changed || export_changed)
+    log(L_INFO "Reloading protocol %s", p->name);
+
+  if (import_changed && ! p->reload_routes(p))
+    {
+      /* Now, the protocol is reconfigured. But route reload failed
+	 and we have to do regular protocol restart. */
+      log(L_INFO "Restarting protocol %s", p->name);
+      p->disabled = 1;
+      proto_rethink_goal(p);
+      p->disabled = 0;
+      proto_rethink_goal(p);
+      return 1;
+    }
+
+  if (export_changed)
+    proto_request_feeding(p);
+
+  return 1;
+}
+
 /**
  * protos_commit - commit new protocol configuration
  * @new: new configuration
@@ -307,40 +371,29 @@ protos_commit(struct config *new, struct config *old, int force_reconfig, int ty
 	  if (sym && sym->class == SYM_PROTO && !new->shutdown)
 	    {
 	      /* Found match, let's check if we can smoothly switch to new configuration */
+	      /* No need to check description */
 	      nc = sym->def;
-	      if (!force_reconfig
-		  && nc->protocol == oc->protocol
-		  && nc->preference == oc->preference
-		  && nc->disabled == oc->disabled
-		  && nc->table->table == oc->table->table
-		  && proto_get_router_id(nc) == proto_get_router_id(oc)
-		  && ((type == RECONFIG_SOFT) || filter_same(nc->in_filter, oc->in_filter))
-		  && ((type == RECONFIG_SOFT) || filter_same(nc->out_filter, oc->out_filter))
-		  && p->proto_state != PS_DOWN)
-		{
-		  /* Generic attributes match, try converting them and then ask the protocol */
-		  p->debug = nc->debug;
-		  if (p->proto->reconfigure && p->proto->reconfigure(p, nc))
-		    {
-		      DBG("\t%s: same\n", oc->name);
-		      PD(p, "Reconfigured");
-		      p->cf = nc;
-		      p->name = nc->name;
-		      p->in_filter = nc->in_filter;
-		      p->out_filter = nc->out_filter;
-		      nc->proto = p;
-		      continue;
-		    }
-		}
-	      /* Unsuccessful, force reconfig */
-	      DBG("\t%s: power cycling\n", oc->name);
-	      PD(p, "Reconfiguration failed, restarting");
-	      p->cf_new = nc;
 	      nc->proto = p;
+
+	      /* We will try to reconfigure protocol p */
+	      if (! force_reconfig && proto_reconfigure(p, oc, nc, type))
+		continue;
+
+	      /* Unsuccessful, we will restart it */
+	      if (!p->disabled && !nc->disabled)
+		log(L_INFO "Restarting protocol %s", p->name);
+	      else if (p->disabled && !nc->disabled)
+		log(L_INFO "Enabling protocol %s", p->name);
+	      else if (!p->disabled && nc->disabled)
+		log(L_INFO "Disabling protocol %s", p->name);
+
+	      PD(p, "Restarting");
+	      p->cf_new = nc;
 	    }
 	  else
 	    {
-	      DBG("\t%s: deleting\n", oc->name);
+	      if (!shutting_down)
+		log(L_INFO "Removing protocol %s", p->name);
 	      PD(p, "Unconfigured");
 	      p->cf_new = NULL;
 	    }
@@ -353,7 +406,8 @@ protos_commit(struct config *new, struct config *old, int force_reconfig, int ty
   WALK_LIST(nc, new->protos)
     if (!nc->proto)
       {
-	DBG("\t%s: adding\n", nc->name);
+	if (old_config)		/* Not a first-time configuration */
+	  log(L_INFO "Adding protocol %s", nc->name);
 	proto_init(nc);
       }
   DBG("\tdone\n");
@@ -522,6 +576,12 @@ proto_fell_down(struct proto *p)
 
   bzero(&p->stats, sizeof(struct proto_stats));
   rt_unlock_table(p->table);
+
+#ifdef CONFIG_PIPE
+  if (proto_is_pipe(p))
+    rt_unlock_table(pipe_get_peer_table(p));
+#endif
+
   proto_rethink_goal(p);
 }
 
@@ -548,7 +608,7 @@ proto_feed_more(void *P)
 }
 
 static void
-proto_feed(void *P)
+proto_feed_initial(void *P)
 {
   struct proto *p = P;
 
@@ -576,15 +636,50 @@ proto_schedule_flush(struct proto *p)
 }
 
 static void
-proto_schedule_feed(struct proto *p)
+proto_schedule_feed(struct proto *p, int initial)
 {
   DBG("%s: Scheduling meal\n", p->name);
   p->core_state = FS_FEEDING;
+  p->refeeding = !initial;
+
+  /* Hack: reset exp_routes during refeed, and do not decrease it later */
+  if (!initial)
+    p->stats.exp_routes = 0;
+
   proto_relink(p);
-  p->attn->hook = proto_feed;
+  p->attn->hook = initial ? proto_feed_initial : proto_feed_more;
   ev_schedule(p->attn);
 }
 
+/**
+ * proto_request_feeding - request feeding routes to the protocol
+ * @p: given protocol 
+ *
+ * Sometimes it is needed to send again all routes to the
+ * protocol. This is called feeding and can be requested by this
+ * function. This would cause protocol core state transition
+ * to FS_FEEDING (during feeding) and when completed, it will
+ * switch back to FS_HAPPY. This function can be called even
+ * when feeding is already running, in that case it is restarted.
+ */
+void
+proto_request_feeding(struct proto *p)
+{
+  ASSERT(p->proto_state == PS_UP);
+
+  /* If we are already feeding, we want to restart it */
+  if (p->core_state == FS_FEEDING)
+    {
+      /* Unless feeding is in initial state */
+      if (p->attn->hook == proto_feed_initial)
+	return;
+
+      rt_feed_baby_abort(p);
+    }
+
+  proto_schedule_feed(p, 0);
+}
+
 /**
  * proto_notify_state - notify core about protocol state change
  * @p: protocol the state of which has changed
@@ -614,7 +709,7 @@ proto_notify_state(struct proto *p, unsigned ps)
   switch (ps)
     {
     case PS_DOWN:
-      if ((cs = FS_FEEDING) || (cs == FS_HAPPY))
+      if ((cs == FS_FEEDING) || (cs == FS_HAPPY))
 	proto_schedule_flush(p);
 
       neigh_prune(); // FIXME convert neighbors to resource?
@@ -634,10 +729,10 @@ proto_notify_state(struct proto *p, unsigned ps)
     case PS_UP:
       ASSERT(ops == PS_DOWN || ops == PS_START);
       ASSERT(cs == FS_HUNGRY);
-      proto_schedule_feed(p);
+      proto_schedule_feed(p, 1);
       break;
     case PS_STOP:
-      if ((cs = FS_FEEDING) || (cs == FS_HAPPY))
+      if ((cs == FS_FEEDING) || (cs == FS_HAPPY))
 	proto_schedule_flush(p);
       break;
     default:
@@ -645,6 +740,8 @@ proto_notify_state(struct proto *p, unsigned ps)
     }
 }
 
+extern struct protocol proto_unix_iface;
+
 static void
 proto_flush_all(void *unused UNUSED)
 {
@@ -653,6 +750,11 @@ proto_flush_all(void *unused UNUSED)
   rt_prune_all();
   while ((p = HEAD(flush_proto_list))->n.next)
     {
+      /* This will flush interfaces in the same manner
+	 like rt_prune_all() flushes routes */
+      if (p->proto == &proto_unix_iface)
+	if_flush_ifaces(p);
+
       DBG("Flushing protocol %s\n", p->name);
       p->core_state = FS_HUNGRY;
       proto_relink(p);
@@ -685,70 +787,250 @@ proto_state_name(struct proto *p)
 }
 
 static void
-proto_do_show(struct proto *p, int verbose)
+proto_do_show_stats(struct proto *p)
 {
-  byte buf[256], reltime[TM_RELTIME_BUFFER_SIZE];
+  struct proto_stats *s = &p->stats;
+  cli_msg(-1006, "  Routes:         %u imported, %u exported, %u preferred", 
+	  s->imp_routes, s->exp_routes, s->pref_routes);
+  cli_msg(-1006, "  Route change stats:     received   rejected   filtered    ignored   accepted");
+  cli_msg(-1006, "    Import updates:     %10u %10u %10u %10u %10u",
+	  s->imp_updates_received, s->imp_updates_invalid,
+	  s->imp_updates_filtered, s->imp_updates_ignored,
+	  s->imp_updates_accepted);
+  cli_msg(-1006, "    Import withdraws:   %10u %10u        --- %10u %10u",
+	  s->imp_withdraws_received, s->imp_withdraws_invalid,
+	  s->imp_withdraws_ignored, s->imp_withdraws_accepted);
+  cli_msg(-1006, "    Export updates:     %10u %10u %10u        --- %10u",
+	  s->exp_updates_received, s->exp_updates_rejected,
+	  s->exp_updates_filtered, s->exp_updates_accepted);
+  cli_msg(-1006, "    Export withdraws:   %10u        ---        ---        --- %10u",
+	  s->exp_withdraws_received, s->exp_withdraws_accepted);
+}
+
+#ifdef CONFIG_PIPE
+static void
+proto_do_show_pipe_stats(struct proto *p)
+{
+  struct proto_stats *s1 = &p->stats;
+  struct proto_stats *s2 = pipe_get_peer_stats(p);
+
+  /*
+   * Pipe stats (as anything related to pipes) are a bit tricky. There
+   * are two sets of stats - s1 for routes going from the primary
+   * routing table to the secondary routing table ('exported' from the
+   * user point of view) and s2 for routes going in the other
+   * direction ('imported' from the user point of view).
+   *
+   * Each route going through a pipe is, technically, first exported
+   * to the pipe and then imported from that pipe and such operations
+   * are counted in one set of stats according to the direction of the
+   * route propagation. Filtering is done just in the first part
+   * (export). Therefore, we compose stats for one directon for one
+   * user direction from both import and export stats, skipping
+   * immediate and irrelevant steps (exp_updates_accepted,
+   * imp_updates_received, imp_updates_filtered, ...)
+   */
+
+  cli_msg(-1006, "  Routes:         %u imported, %u exported", 
+	  s2->imp_routes, s1->imp_routes);
+  cli_msg(-1006, "  Route change stats:     received   rejected   filtered    ignored   accepted");
+  cli_msg(-1006, "    Import updates:     %10u %10u %10u %10u %10u",
+	  s2->exp_updates_received, s2->exp_updates_rejected + s2->imp_updates_invalid,
+	  s2->exp_updates_filtered, s2->imp_updates_ignored, s2->imp_updates_accepted);
+  cli_msg(-1006, "    Import withdraws:   %10u %10u        --- %10u %10u",
+	  s2->exp_withdraws_received, s2->imp_withdraws_invalid,
+	  s2->imp_withdraws_ignored, s2->imp_withdraws_accepted);
+  cli_msg(-1006, "    Export updates:     %10u %10u %10u %10u %10u",
+	  s1->exp_updates_received, s1->exp_updates_rejected + s1->imp_updates_invalid,
+	  s1->exp_updates_filtered, s1->imp_updates_ignored, s1->imp_updates_accepted);
+  cli_msg(-1006, "    Export withdraws:   %10u %10u        --- %10u %10u",
+	  s1->exp_withdraws_received, s1->imp_withdraws_invalid,
+	  s1->imp_withdraws_ignored, s1->imp_withdraws_accepted);
+}
+#endif
+
+void
+proto_cmd_show(struct proto *p, unsigned int verbose, int cnt)
+{
+  byte buf[256], tbuf[TM_DATETIME_BUFFER_SIZE];
+
+  /* First protocol - show header */
+  if (!cnt)
+    cli_msg(-2002, "name     proto    table    state  since       info");
 
   buf[0] = 0;
   if (p->proto->get_status)
     p->proto->get_status(p, buf);
-  tm_format_reltime(reltime, p->last_state_change);
-  cli_msg(-1002, "%-8s %-8s %-8s %-5s %-5s  %s",
+  tm_format_datetime(tbuf, &config->tf_proto, p->last_state_change);
+  cli_msg(-1002, "%-8s %-8s %-8s %-5s  %-10s  %s",
 	  p->name,
 	  p->proto->name,
 	  p->table->name,
 	  proto_state_name(p),
-	  reltime,
+	  tbuf,
 	  buf);
   if (verbose)
     {
+      if (p->cf->dsc)
+	cli_msg(-1006, "  Description:    %s", p->cf->dsc);
+      if (p->cf->router_id)
+	cli_msg(-1006, "  Router ID:      %R", p->cf->router_id);
       cli_msg(-1006, "  Preference:     %d", p->preference);
       cli_msg(-1006, "  Input filter:   %s", filter_name(p->in_filter));
       cli_msg(-1006, "  Output filter:  %s", filter_name(p->out_filter));
 
       if (p->proto_state != PS_DOWN)
 	{
-	  cli_msg(-1006, "  Routes:         %u imported, %u exported, %u preferred", 
-		  p->stats.imp_routes, p->stats.exp_routes, p->stats.pref_routes);
-	  cli_msg(-1006, "  Route change stats:     received   rejected   filtered    ignored   accepted");
-	  cli_msg(-1006, "    Import updates:     %10u %10u %10u %10u %10u",
-		  p->stats.imp_updates_received, p->stats.imp_updates_invalid,
-		  p->stats.imp_updates_filtered, p->stats.imp_updates_ignored,
-		  p->stats.imp_updates_accepted);
-	  cli_msg(-1006, "    Import withdraws:   %10u %10u        --- %10u %10u",
-		  p->stats.imp_withdraws_received, p->stats.imp_withdraws_invalid,
-		  p->stats.imp_withdraws_ignored, p->stats.imp_withdraws_accepted);
-	  cli_msg(-1006, "    Export updates:     %10u %10u %10u        --- %10u",
-		  p->stats.exp_updates_received, p->stats.exp_updates_rejected,
-		  p->stats.exp_updates_filtered, p->stats.exp_updates_accepted);
-	  cli_msg(-1006, "    Export withdraws:   %10u        ---        ---        --- %10u",
-		  p->stats.exp_withdraws_received, p->stats.exp_withdraws_accepted);
+#ifdef CONFIG_PIPE
+	  if (proto_is_pipe(p))
+	    proto_do_show_pipe_stats(p);
+	  else
+#endif
+	    proto_do_show_stats(p);
 	}
 
+      if (p->proto->show_proto_info)
+	p->proto->show_proto_info(p);
+
       cli_msg(-1006, "");
     }
 }
 
 void
-proto_show(struct symbol *s, int verbose)
+proto_cmd_disable(struct proto *p, unsigned int arg UNUSED, int cnt UNUSED)
 {
-  if (s && s->class != SYM_PROTO)
+  if (p->disabled)
+    {
+      cli_msg(-8, "%s: already disabled", p->name);
+      return;
+    }
+
+  log(L_INFO "Disabling protocol %s", p->name);
+  p->disabled = 1;
+  proto_rethink_goal(p);
+  cli_msg(-9, "%s: disabled", p->name);
+}
+
+void
+proto_cmd_enable(struct proto *p, unsigned int arg UNUSED, int cnt UNUSED)
+{
+  if (!p->disabled)
+    {
+      cli_msg(-10, "%s: already enabled", p->name);
+      return;
+    }
+
+  log(L_INFO "Enabling protocol %s", p->name);
+  p->disabled = 0;
+  proto_rethink_goal(p);
+  cli_msg(-11, "%s: enabled", p->name);
+}
+
+void
+proto_cmd_restart(struct proto *p, unsigned int arg UNUSED, int cnt UNUSED)
+{
+  if (p->disabled)
+    {
+      cli_msg(-8, "%s: already disabled", p->name);
+      return;
+    }
+
+  log(L_INFO "Restarting protocol %s", p->name);
+  p->disabled = 1;
+  proto_rethink_goal(p);
+  p->disabled = 0;
+  proto_rethink_goal(p);
+  cli_msg(-12, "%s: restarted", p->name);
+}
+
+void
+proto_cmd_reload(struct proto *p, unsigned int dir, int cnt UNUSED)
+{
+  if (p->disabled)
+    {
+      cli_msg(-8, "%s: already disabled", p->name);
+      return;
+    }
+
+  /* If the protocol in not UP, it has no routes */
+  if (p->proto_state != PS_UP)
+    return;
+
+  log(L_INFO "Reloading protocol %s", p->name);
+
+  /* re-importing routes */
+  if (dir != CMD_RELOAD_OUT)
+    if (! (p->reload_routes && p->reload_routes(p)))
+      {
+	cli_msg(-8006, "%s: reload failed", p->name);
+	return;
+      }
+		 
+  /* re-exporting routes */
+  if (dir != CMD_RELOAD_IN)
+    proto_request_feeding(p);
+
+  cli_msg(-15, "%s: reloading", p->name);
+}
+
+void
+proto_cmd_debug(struct proto *p, unsigned int mask, int cnt UNUSED)
+{
+  p->debug = mask;
+}
+
+void
+proto_cmd_mrtdump(struct proto *p, unsigned int mask, int cnt UNUSED)
+{
+  p->mrtdump = mask;
+}
+
+static void
+proto_apply_cmd_symbol(struct symbol *s, void (* cmd)(struct proto *, unsigned int, int), unsigned int arg)
+{
+  if (s->class != SYM_PROTO)
     {
       cli_msg(9002, "%s is not a protocol", s->name);
       return;
     }
-  cli_msg(-2002, "name     proto    table    state since  info");
-  if (s)
-    proto_do_show(((struct proto_config *)s->def)->proto, verbose);
-  else
-    {
-      WALK_PROTO_LIST(p)
-	proto_do_show(p, verbose);
-      WALK_PROTO_LIST_END;
-    }
+
+  cmd(((struct proto_config *)s->def)->proto, arg, 0);
   cli_msg(0, "");
 }
 
+static void
+proto_apply_cmd_patt(char *patt, void (* cmd)(struct proto *, unsigned int, int), unsigned int arg)
+{
+  int cnt = 0;
+
+  node *nn;
+  WALK_LIST(nn, proto_list)
+    {
+      struct proto *p = SKIP_BACK(struct proto, glob_node, nn);
+
+      if (!patt || patmatch(patt, p->name))
+	cmd(p, arg, cnt++);
+    }
+
+  if (!cnt)
+    cli_msg(8003, "No protocols match");
+  else
+    cli_msg(0, "");
+}
+
+void
+proto_apply_cmd(struct proto_spec ps, void (* cmd)(struct proto *, unsigned int, int),
+		int restricted, unsigned int arg)
+{
+  if (restricted && cli_access_restricted())
+    return;
+
+  if (ps.patt)
+    proto_apply_cmd_patt(ps.ptr, cmd, arg);
+  else
+    proto_apply_cmd_symbol(ps.ptr, cmd, arg);
+}
+
 struct proto *
 proto_get_named(struct symbol *sym, struct protocol *pr)
 {
@@ -777,71 +1059,3 @@ proto_get_named(struct symbol *sym, struct protocol *pr)
     }
   return p;
 }
-
-void
-proto_xxable(char *pattern, int xx)
-{
-  int cnt = 0;
-  WALK_PROTO_LIST(p)
-    if (patmatch(pattern, p->name))
-      {
-	cnt++;
-	switch (xx)
-	  {
-	  case 0:
-	    if (p->disabled)
-	      cli_msg(-8, "%s: already disabled", p->name);
-	    else
-	      {
-		cli_msg(-9, "%s: disabled", p->name);
-		p->disabled = 1;
-	      }
-	    break;
-	  case 1:
-	    if (!p->disabled)
-	      cli_msg(-10, "%s: already enabled", p->name);
-	    else
-	      {
-		cli_msg(-11, "%s: enabled", p->name);
-		p->disabled = 0;
-	      }
-	    break;
-	  case 2:
-	    if (p->disabled)
-	      cli_msg(-8, "%s: already disabled", p->name);
-	    else
-	      {
-		p->disabled = 1;
-		proto_rethink_goal(p);
-		p->disabled = 0;
-		cli_msg(-12, "%s: restarted", p->name);
-	      }
-	    break;
-	  default:
-	    ASSERT(0);
-	  }
-	proto_rethink_goal(p);
-      }
-  WALK_PROTO_LIST_END;
-  if (!cnt)
-    cli_msg(8003, "No protocols match");
-  else
-    cli_msg(0, "");
-}
-
-void
-proto_debug(char *pattern, unsigned int mask)
-{
-  int cnt = 0;
-  WALK_PROTO_LIST(p)
-    if (patmatch(pattern, p->name))
-      {
-	cnt++;
-	p->debug = mask;
-      }
-  WALK_PROTO_LIST_END;
-  if (!cnt)
-    cli_msg(8003, "No protocols match");
-  else
-    cli_msg(0, "");
-}
