@@ -42,6 +42,32 @@
  * and deletion. Each LSA is kept in two pieces: header and body. Both of them are
  * kept in the endianity of the CPU.
  *
+ * In OSPFv2 specification, it is implied that there is one IP prefix
+ * for each physical network/interface (unless it is an ptp link). But
+ * in modern systems, there might be more independent IP prefixes
+ * associated with an interface.  To handle this situation, we have
+ * one &ospf_iface for each active IP prefix (instead for each active
+ * iface); This behaves like virtual interface for the purpose of OSPF.
+ * If we receive packet, we associate it with a proper virtual interface
+ * mainly according to its source address.
+ *
+ * OSPF keeps one socket per &ospf_iface. This allows us (compared to
+ * one socket approach) to evade problems with a limit of multicast
+ * groups per socket and with sending multicast packets to appropriate
+ * interface in a portable way. The socket is associated with
+ * underlying physical iface and should not receive packets received
+ * on other ifaces (unfortunately, this is not true on
+ * BSD). Generally, one packet can be received by more sockets (for
+ * example, if there are more &ospf_iface on one physical iface),
+ * therefore we explicitly filter received packets according to
+ * src/dst IP address and received iface.
+ *
+ * Vlinks are implemented using particularly degenerate form of
+ * &ospf_iface, which has several exceptions: it does not have its
+ * iface or socket (it copies these from 'parent' &ospf_iface) and it
+ * is present in iface list even when down (it is not freed in
+ * ospf_iface_down()).
+ *
  * The heart beat of ospf is ospf_disp(). It is called at regular intervals
  * (&proto_ospf->tick). It is responsible for aging and flushing of LSAs in
  * the database, for routing table calculaction and it call area_disp() of every
@@ -49,7 +75,7 @@
  * 
  * The function area_disp() is
  * responsible for late originating of router LSA and network LSA
- * and for cleanup after routing table calculation process in
+ * and for cleanup before routing table calculation process in
  * the area.
  * To every &ospf_iface, we connect one or more
  * &ospf_neighbor's -- a structure containing many timers and queues
@@ -77,8 +103,8 @@
 #include "ospf.h"
 
 
-static void ospf_rt_notify(struct proto *p, net * n, rte * new, rte * old UNUSED, ea_list * attrs);
-static void ospf_ifa_notify(struct proto *p, unsigned flags, struct ifa *a);
+static int ospf_reload_routes(struct proto *p);
+static void ospf_rt_notify(struct proto *p, struct rtable *table UNUSED, net * n, rte * new, rte * old UNUSED, ea_list * attrs);
 static int ospf_rte_better(struct rte *new, struct rte *old);
 static int ospf_rte_same(struct rte *new, struct rte *old);
 static void ospf_disp(timer *timer);
@@ -116,10 +142,11 @@ ospf_start(struct proto *p)
   struct ospf_config *c = (struct ospf_config *) (p->cf);
   struct ospf_area_config *ac;
   struct ospf_area *oa;
+  int vlinks = 0;
 
+  po->router_id = proto_get_router_id(p->cf);
   po->rfc1583 = c->rfc1583;
   po->ebit = 0;
-
   po->tick = c->tick;
   po->disp_timer = tm_new(p->pool);
   po->disp_timer->data = po;
@@ -135,7 +162,6 @@ ospf_start(struct proto *p)
   fib_init(&po->rtf, p->pool, sizeof(ort), 16, ospf_rt_initort);
   po->areano = 0;
   po->gr = ospf_top_new(p->pool);
-  po->cleanup = 1;
   s_init_list(&(po->lsal));
   if (EMPTY_LIST(c->area_list))
   {
@@ -163,36 +189,42 @@ ospf_start(struct proto *p)
       oa->stub = 0;
     }
 
-    oa->opt.byte = 0;
-    if(!oa->stub) oa->opt.bit.e = 1;
+    if (!EMPTY_LIST(ac->vlink_list))
+      vlinks = 1;
+
+#ifdef OSPFv2
+    oa->options = (oa->stub ? 0 : OPT_E);
+#else /* OSPFv3 */
+    oa->options = OPT_R | (oa->stub ? 0 : OPT_E) | OPT_V6;
+#endif
+  }
+
+  /* ABR is always in the backbone */
+  if (((po->areano > 1) || vlinks) && !po->backbone)
+  {
+    oa = mb_allocz(p->pool, sizeof(struct ospf_area));
+    add_tail(&po->area_list, NODE oa);
+    po->areano++;
+    oa->stub = 0;
+    oa->areaid = 0;
+    oa->rt = NULL;
+    oa->po = po;
+    fib_init(&oa->net_fib, p->pool, sizeof(struct area_net), 16, ospf_area_initfib);
+    fib_init(&oa->rtr, p->pool, sizeof(ort), 16, ospf_rt_initort);
+    po->backbone = oa;
+#ifdef OSPFv2
+    oa->options = OPT_E;
+#else /* OSPFv3 */
+    oa->options = OPT_R | OPT_E | OPT_V6;
+#endif
   }
 
   /* Add all virtual links as interfaces */
-  {
-    struct ospf_iface_patt *ipatt;
-    WALK_LIST(ac, c->area_list)
-    {
-      WALK_LIST(ipatt, ac->vlink_list)
-      {
-        if(!po->backbone)
-	{
-          oa = mb_allocz(p->pool, sizeof(struct ospf_area));
-          add_tail(&po->area_list, NODE oa);
-          po->areano++;
-          oa->stub = 0;
-          oa->areaid = 0;
-          oa->rt = NULL;
-          oa->po = po;
-	  fib_init(&oa->net_fib, p->pool, sizeof(struct area_net), 16, ospf_area_initfib);
-          fib_init(&oa->rtr, p->pool, sizeof(ort), 16, ospf_rt_initort);
-          po->backbone = oa;
-          oa->opt.byte = 0;
-          oa->opt.bit.e = 1;
-	}
-        ospf_iface_new(po, NULL, ac, ipatt);
-      }
-    }
-  }
+  struct ospf_iface_patt *ipatt;
+  WALK_LIST(ac, c->area_list)
+    WALK_LIST(ipatt, ac->vlink_list)
+    ospf_iface_new(po, NULL, NULL, ac, ipatt);
+
   return PS_UP;
 }
 
@@ -217,9 +249,11 @@ ospf_dump(struct proto *p)
     }
   }
 
+  /*
   OSPF_TRACE(D_EVENTS, "LSA graph dump start:");
   ospf_top_dump(po->gr, p);
   OSPF_TRACE(D_EVENTS, "LSA graph dump finished");
+  */
   neigh_dump_all();
 }
 
@@ -228,9 +262,10 @@ ospf_init(struct proto_config *c)
 {
   struct proto *p = proto_new(c, sizeof(struct proto_ospf));
 
-  p->import_control = ospf_import_control;
   p->make_tmp_attrs = ospf_make_tmp_attrs;
   p->store_tmp_attrs = ospf_store_tmp_attrs;
+  p->import_control = ospf_import_control;
+  p->reload_routes = ospf_reload_routes;
   p->accept_ra_types = RA_OPTIMAL;
   p->rt_notify = ospf_rt_notify;
   p->if_notify = ospf_iface_notify;
@@ -270,19 +305,20 @@ ospf_rte_same(struct rte *new, struct rte *old)
   return
     new->u.ospf.metric1 == old->u.ospf.metric1 &&
     new->u.ospf.metric2 == old->u.ospf.metric2 &&
-    new->u.ospf.tag == old->u.ospf.tag;
+    new->u.ospf.tag == old->u.ospf.tag &&
+    new->u.ospf.router_id == old->u.ospf.router_id;
 }
 
 static ea_list *
 ospf_build_attrs(ea_list * next, struct linpool *pool, u32 m1, u32 m2,
-		 u32 tag)
+		 u32 tag, u32 rid)
 {
   struct ea_list *l =
-    lp_alloc(pool, sizeof(struct ea_list) + 3 * sizeof(eattr));
+    lp_alloc(pool, sizeof(struct ea_list) + 4 * sizeof(eattr));
 
   l->next = next;
   l->flags = EALF_SORTED;
-  l->count = 3;
+  l->count = 4;
   l->attrs[0].id = EA_OSPF_METRIC1;
   l->attrs[0].flags = 0;
   l->attrs[0].type = EAF_TYPE_INT | EAF_TEMP;
@@ -295,21 +331,39 @@ ospf_build_attrs(ea_list * next, struct linpool *pool, u32 m1, u32 m2,
   l->attrs[2].flags = 0;
   l->attrs[2].type = EAF_TYPE_INT | EAF_TEMP;
   l->attrs[2].u.data = tag;
+  l->attrs[3].id = EA_OSPF_ROUTER_ID;
+  l->attrs[3].flags = 0;
+  l->attrs[3].type = EAF_TYPE_ROUTER_ID | EAF_TEMP;
+  l->attrs[3].u.data = rid;
   return l;
 }
 
 void
 schedule_net_lsa(struct ospf_iface *ifa)
 {
+  struct proto *p = &ifa->oa->po->proto;
+
+  OSPF_TRACE(D_EVENTS, "Scheduling network-LSA origination for iface %s", ifa->iface->name);
   ifa->orignet = 1;
 }
 
+#ifdef OSPFv3
+void
+schedule_link_lsa(struct ospf_iface *ifa)
+{
+  struct proto *p = &ifa->oa->po->proto;
+
+  OSPF_TRACE(D_EVENTS, "Scheduling link-LSA origination for iface %s", ifa->iface->name);
+  ifa->origlink = 1;
+}
+#endif
+
 void
 schedule_rt_lsa(struct ospf_area *oa)
 {
   struct proto *p = &oa->po->proto;
 
-  OSPF_TRACE(D_EVENTS, "Scheduling RT lsa origination for area %R.", oa->areaid);
+  OSPF_TRACE(D_EVENTS, "Scheduling router-LSA origination for area %R", oa->areaid);
   oa->origrt = 1;
 }
 
@@ -321,17 +375,30 @@ schedule_rtcalc(struct proto_ospf *po)
   if (po->calcrt)
     return;
 
-  OSPF_TRACE(D_EVENTS, "Scheduling RT calculation.");
+  OSPF_TRACE(D_EVENTS, "Scheduling routing table calculation");
   po->calcrt = 1;
 }
 
+static int
+ospf_reload_routes(struct proto *p)
+{
+  struct proto_ospf *po = (struct proto_ospf *) p;
+
+  if (po->calcrt != 2)
+    OSPF_TRACE(D_EVENTS, "Scheduling routing table calculation with route reload");
+
+  po->calcrt = 2;
+
+  return 1;
+}
+
 /**
  * area_disp - invokes origination of
  * router LSA and routing table cleanup
  * @oa: ospf area
  *
  * It invokes aging and when @ospf_area->origrt is set to 1, start
- * function for origination of router LSA and network LSAs.
+ * function for origination of router, network LSAs.
  */
 void
 area_disp(struct ospf_area *oa)
@@ -341,13 +408,19 @@ area_disp(struct ospf_area *oa)
 
   /* Now try to originage rt_lsa */
   if (oa->origrt)
-    originate_rt_lsa(oa);
+    update_rt_lsa(oa);
 
   /* Now try to originate network LSA's */
   WALK_LIST(ifa, po->iface_list)
   {
+#ifdef OSPFv3
+    /* Link LSA should be originated before Network LSA */
+    if (ifa->origlink && (ifa->oa == oa))
+      update_link_lsa(ifa);
+#endif
+
     if (ifa->orignet && (ifa->oa == oa))
-      originate_net_lsa(ifa);
+      update_net_lsa(ifa);
   }
 }
 
@@ -370,7 +443,7 @@ ospf_disp(timer * timer)
 
   /* Calculate routing table */
   if (po->calcrt)
-    ospf_rt_spf (po);
+    ospf_rt_spf(po);
 }
 
 
@@ -394,7 +467,7 @@ ospf_import_control(struct proto *p, rte ** new, ea_list ** attrs,
 
   if (p == e->attrs->proto)
     return -1;			/* Reject our own routes */
-  *attrs = ospf_build_attrs(*attrs, pool, LSINFINITY, 10000, 0);
+  *attrs = ospf_build_attrs(*attrs, pool, LSINFINITY, 10000, 0, 0);
   return 0;			/* Leave decision to the filters */
 }
 
@@ -402,7 +475,7 @@ struct ea_list *
 ospf_make_tmp_attrs(struct rte *rt, struct linpool *pool)
 {
   return ospf_build_attrs(NULL, pool, rt->u.ospf.metric1, rt->u.ospf.metric2,
-			  rt->u.ospf.tag);
+			  rt->u.ospf.tag, rt->u.ospf.router_id);
 }
 
 void
@@ -411,6 +484,7 @@ ospf_store_tmp_attrs(struct rte *rt, struct ea_list *attrs)
   rt->u.ospf.metric1 = ea_get_int(attrs, EA_OSPF_METRIC1, LSINFINITY);
   rt->u.ospf.metric2 = ea_get_int(attrs, EA_OSPF_METRIC2, 10000);
   rt->u.ospf.tag = ea_get_int(attrs, EA_OSPF_TAG, 0);
+  rt->u.ospf.router_id = ea_get_int(attrs, EA_OSPF_ROUTER_ID, 0);
 }
 
 /**
@@ -431,14 +505,15 @@ ospf_shutdown(struct proto *p)
   OSPF_TRACE(D_EVENTS, "Shutdown requested");
 
   /* And send to all my neighbors 1WAY */
-  WALK_LIST(ifa, po->iface_list) ospf_iface_shutdown(ifa);
+  WALK_LIST(ifa, po->iface_list)
+    if (ifa->state > OSPF_IS_DOWN)
+      ospf_iface_shutdown(ifa);
 
   return PS_DOWN;
 }
 
 static void
-ospf_rt_notify(struct proto *p, net * n, rte * new, rte * old UNUSED,
-	       ea_list * attrs)
+ospf_rt_notify(struct proto *p, rtable *tbl UNUSED, net * n, rte * new, rte * old UNUSED, ea_list * attrs)
 {
   struct proto_ospf *po = (struct proto_ospf *) p;
 
@@ -448,55 +523,9 @@ ospf_rt_notify(struct proto *p, net * n, rte * new, rte * old UNUSED,
 */
 
   if (new)			/* Got some new route */
-  {
     originate_ext_lsa(n, new, po, attrs);
-  }
   else
-  {
-    u32 rtid = po->proto.cf->global->router_id;
-    struct ospf_area *oa;
-    struct top_hash_entry *en;
-    u32 pr = ipa_to_u32(n->n.prefix);
-    struct ospf_lsa_ext *ext;
-    int i;
-    int max = max_ext_lsa(n->n.pxlen);
-
-    /* Flush old external LSA */
-    for (i = 0; i < max; i++, pr++)
-    {
-      if (en = ospf_hash_find(po->gr, 0, pr, rtid, LSA_T_EXT))
-      {
-        ext = en->lsa_body;
-	if (ipa_compare(ext->netmask, ipa_mkmask(n->n.pxlen)) == 0)
-	{
-          WALK_LIST(oa, po->area_list)
-          {
-	    ospf_lsupd_flush_nlsa(en, oa);
-	  }
-	}
-        break;
-      }
-    }
-  }
-}
-
-static void
-ospf_ifa_notify(struct proto *p, unsigned flags, struct ifa *a)
-{
-  struct proto_ospf *po = (struct proto_ospf *) p;
-  struct ospf_iface *ifa;
-  
-  if ((a->flags & IA_SECONDARY) || (a->flags & IA_UNNUMBERED))
-    return;
-
-  WALK_LIST(ifa, po->iface_list)
-    {
-      if (ifa->iface == a->iface)
-	{
-	  schedule_rt_lsa(ifa->oa);
-	  return;
-	}
-    }
+    flush_ext_lsa(n, po);
 }
 
 static void
@@ -553,6 +582,8 @@ ospf_get_route_info(rte * rte, byte * buf, ea_list * attrs UNUSED)
   {
     buf += bsprintf(buf, " [%x]", rte->u.ospf.tag);
   }
+  if (rte->u.ospf.router_id)
+    buf += bsprintf(buf, " [%R]", rte->u.ospf.router_id);
 }
 
 static int
@@ -567,8 +598,11 @@ ospf_get_attr(eattr * a, byte * buf, int buflen UNUSED)
     bsprintf(buf, "metric2");
     return GA_NAME;
   case EA_OSPF_TAG:
-    bsprintf(buf, "tag: %08x", a->u.data);
+    bsprintf(buf, "tag: 0x%08x", a->u.data);
     return GA_FULL;
+  case EA_OSPF_ROUTER_ID:
+    bsprintf(buf, "router_id");
+    return GA_NAME;
   default:
     return GA_UNKNOWN;
   }
@@ -602,10 +636,10 @@ ospf_reconfigure(struct proto *p, struct proto_config *c)
   struct nbma_node *nb1, *nb2, *nbnx;
   struct ospf_area *oa = NULL;
   int found, olddead, newdead;
-  struct area_net_config *anc;
-  struct area_net *an;
+  
+  if (po->rfc1583 != new->rfc1583)
+    return 0;
 
-  po->rfc1583 = new->rfc1583;
   schedule_rtcalc(po);
 
   po->tick = new->tick;
@@ -655,26 +689,8 @@ ospf_reconfigure(struct proto *p, struct proto_config *c)
       schedule_rt_lsa(oa);
 
     /* Change net_list */
-    FIB_WALK(&oa->net_fib, nf)	/* First check if some networks are deleted */
-    {
-      found = 0;
-      WALK_LIST(anc, newac->net_list)
-      {
-        if (ipa_equal(anc->px.addr, nf->prefix) && (anc->px.len == nf->pxlen))
-	{
-	  found = 1;
-	  break;
-	}
-	if (!found) flush_sum_lsa(oa, nf, ORT_NET);	/* And flush them */
-      }
-    }
-    FIB_WALK_END;
-
-    WALK_LIST(anc, newac->net_list)	/* Second add new networks */
-    {
-      an = fib_get(&oa->net_fib, &anc->px.addr, anc->px.len);
-      an->hidden = anc->hidden;
-    }
+    fib_free(&oa->net_fib);
+    add_area_nets(oa, newac);
 
     if (!iface_patts_equal(&oldac->patt_list, &newac->patt_list,
 			   (void *) ospf_patt_compare))
@@ -682,12 +698,17 @@ ospf_reconfigure(struct proto *p, struct proto_config *c)
 
     WALK_LIST(ifa, po->iface_list)
     {
+      /* FIXME: better handling of vlinks */
+      if (ifa->iface == NULL)
+        continue;
+
+      /* FIXME: better matching of interface_id in OSPFv3 */
       if (oldip = (struct ospf_iface_patt *)
-	  iface_patt_find(&oldac->patt_list, ifa->iface))
+	  iface_patt_find(&oldac->patt_list, ifa->iface, ifa->addr))
       {
 	/* Now reconfigure interface */
 	if (!(newip = (struct ospf_iface_patt *)
-	      iface_patt_find(&newac->patt_list, ifa->iface)))
+	      iface_patt_find(&newac->patt_list, ifa->iface, ifa->addr)))
 	  return 0;
 
 	/* HELLO TIMER */
@@ -748,20 +769,20 @@ ospf_reconfigure(struct proto *p, struct proto_config *c)
 	}
 
 	/* stub */
-	if ((oldip->stub == 0) && (newip->stub != 0))
+	int old_stub = ospf_iface_stubby(oldip, ifa->addr);
+	int new_stub = ospf_iface_stubby(newip, ifa->addr);
+	if (!old_stub && new_stub)
 	{
-	  ifa->stub = newip->stub;
+	  ifa->stub = 1;
 	  OSPF_TRACE(D_EVENTS, "Interface %s is now stub.", ifa->iface->name);
 	}
-	if ((oldip->stub != 0) && (newip->stub == 0) &&
-	    ((ifa->ioprob & OSPF_I_IP) == 0) &&
-	    (((ifa->ioprob & OSPF_I_MC) == 0) || (ifa->type == OSPF_IT_NBMA)))
+	if (old_stub && !new_stub && (ifa->ioprob == OSPF_I_OK))
 	{
-	  ifa->stub = newip->stub;
-	  OSPF_TRACE(D_EVENTS,
-		     "Interface %s is no longer stub.", ifa->iface->name);
+	  ifa->stub = 0;
+	  OSPF_TRACE(D_EVENTS, "Interface %s is no longer stub.", ifa->iface->name);
 	}
 
+#ifdef OSPFv2	
 	/* AUTHENTICATION */
 	if (oldip->autype != newip->autype)
 	{
@@ -772,6 +793,7 @@ ospf_reconfigure(struct proto *p, struct proto_config *c)
 	}
         /* Add *passwords */
 	ifa->passwords = newip->passwords;
+#endif
 
 	/* priority */
 	if (oldip->priority != newip->priority)
@@ -885,7 +907,7 @@ ospf_reconfigure(struct proto *p, struct proto_config *c)
 void
 ospf_sh_neigh(struct proto *p, char *iff)
 {
-  struct ospf_iface *ifa = NULL, *f;
+  struct ospf_iface *ifa = NULL;
   struct ospf_neighbor *n;
   struct proto_ospf *po = (struct proto_ospf *) p;
 
@@ -896,34 +918,13 @@ ospf_sh_neigh(struct proto *p, char *iff)
     return;
   }
 
-  if (iff != NULL)
-  {
-    WALK_LIST(f, po->iface_list)
-    {
-      if (strcmp(iff, f->iface->name) == 0)
-      {
-	ifa = f;
-	break;
-      }
-    }
-    if (ifa == NULL)
-    {
-      cli_msg(0, "");
-      return;
-    }
-    cli_msg(-1013, "%s:", p->name);
-    cli_msg(-1013, "%-12s\t%3s\t%-15s\t%-5s\t%-12s\t%-10s", "Router ID",
-	    "Pri", "     State", "DTime", "Router IP", "Interface");
-    WALK_LIST(n, ifa->neigh_list) ospf_sh_neigh_info(n);
-    cli_msg(0, "");
-    return;
-  }
-
   cli_msg(-1013, "%s:", p->name);
-  cli_msg(-1013, "%-12s\t%3s\t%-15s\t%-5s\t%-12s\t%-10s", "Router ID", "Pri",
-	  "     State", "DTime", "Router IP", "Interface");
+  cli_msg(-1013, "%-12s\t%3s\t%-15s\t%-5s\t%-10s %-12s", "Router ID", "Pri",
+	  "     State", "DTime", "Interface", "Router IP");
   WALK_LIST(ifa, po->iface_list)
-    WALK_LIST(n, ifa->neigh_list) ospf_sh_neigh_info(n);
+    if ((iff == NULL) || patmatch(iff, ifa->iface->name))
+      WALK_LIST(n, ifa->neigh_list)
+	ospf_sh_neigh_info(n);
   cli_msg(0, "");
 }
 
@@ -997,7 +998,7 @@ void
 ospf_sh_iface(struct proto *p, char *iff)
 {
   struct proto_ospf *po = (struct proto_ospf *) p;
-  struct ospf_iface *ifa = NULL, *f;
+  struct ospf_iface *ifa = NULL;
 
   if (p->proto_state != PS_UP)
   {
@@ -1006,111 +1007,196 @@ ospf_sh_iface(struct proto *p, char *iff)
     return;
   }
 
-  if (iff != NULL)
-  {
-    WALK_LIST(f, po->iface_list)
-    {
-      if (strcmp(iff, f->iface->name) == 0)
-      {
-	ifa = f;
-	break;
-      }
-    }
-
-    if (ifa == NULL)
-    {
-      cli_msg(0, "");
-      return;
-    }
-    cli_msg(-1015, "%s:", p->name);
-    ospf_iface_info(ifa);
-    cli_msg(0, "");
-    return;
-  }
   cli_msg(-1015, "%s:", p->name);
-  WALK_LIST(ifa, po->iface_list) ospf_iface_info(ifa);
+  WALK_LIST(ifa, po->iface_list)
+    if ((iff == NULL) || patmatch(iff, ifa->iface->name))
+      ospf_iface_info(ifa);
   cli_msg(0, "");
 }
 
-/* First we want to separate network-LSAs and other LSAs (because network-LSAs
+/* lsa_compare_for_state() - Compare function for 'show ospf state'
+ *
+ * First we want to separate network-LSAs and other LSAs (because network-LSAs
  * will be presented as network nodes and other LSAs together as router nodes)
  * Network-LSAs are sorted according to network prefix, other LSAs are sorted
  * according to originating router id (to get all LSA needed to represent one
  * router node together). Then, according to LSA type, ID and age.
+ *
+ * For OSPFv3, we have to handle also Prefix-LSAs. We would like to put each
+ * immediately after the referenced LSA. We will make faked LSA based on ref_
+ * values
  */
+
+#ifdef OSPFv3
+
+static struct ospf_lsa_header *
+fake_lsa_from_prefix_lsa(struct ospf_lsa_header *dst, struct ospf_lsa_header *src,
+			 struct ospf_lsa_prefix *px)
+{
+  dst->age = src->age;
+  dst->type = px->ref_type;
+  dst->id = px->ref_id;
+  dst->rt = px->ref_rt;
+  dst->sn = src->sn;
+
+  return dst;
+}
+
+#endif
+
 static int
-he_compare(const void *p1, const void *p2)
+lsa_compare_for_state(const void *p1, const void *p2)
 {
   struct top_hash_entry * he1 = * (struct top_hash_entry **) p1;
   struct top_hash_entry * he2 = * (struct top_hash_entry **) p2;
   struct ospf_lsa_header *lsa1 = &(he1->lsa);
   struct ospf_lsa_header *lsa2 = &(he2->lsa);
+
+  if (he1->domain != he2->domain)
+    return he1->domain - he2->domain;
+
+#ifdef OSPFv3
+  struct ospf_lsa_header lsatmp1, lsatmp2;
+
+  int px1 = (lsa1->type == LSA_T_PREFIX);
+  int px2 = (lsa2->type == LSA_T_PREFIX);
+
+  if (px1)
+    lsa1 = fake_lsa_from_prefix_lsa(&lsatmp1, lsa1, he1->lsa_body);
+
+  if (px2)
+    lsa2 = fake_lsa_from_prefix_lsa(&lsatmp2, lsa2, he2->lsa_body);
+#endif
+
   int nt1 = (lsa1->type == LSA_T_NET);
   int nt2 = (lsa2->type == LSA_T_NET);
 
-  if (he1->oa->areaid != he2->oa->areaid)
-    return he1->oa->areaid - he2->oa->areaid;
-
   if (nt1 != nt2)
     return nt1 - nt2;
 
   if (nt1)
   {
-    // we are cheating for now
+#ifdef OSPFv3
+    /* In OSPFv3, neworks are named base on ID of DR */
+    if (lsa1->rt != lsa2->rt)
+      return lsa1->rt - lsa2->rt;
+#endif
+
+    /* For OSPFv2, this is IP of the network,
+       for OSPFv3, this is interface ID */
     if (lsa1->id != lsa2->id)
       return lsa1->id - lsa2->id;
-    
-    return lsa1->age - lsa2->age;
+
+#ifdef OSPFv3
+    if (px1 != px2)
+      return px1 - px2;
+#endif
+
+    return lsa1->sn - lsa2->sn;
   }
   else 
-    {
-      if (lsa1->rt != lsa2->rt)
-	return lsa1->rt - lsa2->rt;
+  {
+    if (lsa1->rt != lsa2->rt)
+      return lsa1->rt - lsa2->rt;
+
+    if (lsa1->type != lsa2->type)
+      return lsa1->type - lsa2->type;
+  
+    if (lsa1->id != lsa2->id)
+      return lsa1->id - lsa2->id;
+
+#ifdef OSPFv3
+    if (px1 != px2)
+      return px1 - px2;
+#endif
+  
+    return lsa1->sn - lsa2->sn;
+  }
+}
+
+static int
+ext_compare_for_state(const void *p1, const void *p2)
+{
+  struct top_hash_entry * he1 = * (struct top_hash_entry **) p1;
+  struct top_hash_entry * he2 = * (struct top_hash_entry **) p2;
+  struct ospf_lsa_header *lsa1 = &(he1->lsa);
+  struct ospf_lsa_header *lsa2 = &(he2->lsa);
+
+  if (lsa1->rt != lsa2->rt)
+    return lsa1->rt - lsa2->rt;
+
+  if (lsa1->id != lsa2->id)
+    return lsa1->id - lsa2->id;
  
-      if (lsa1->type != lsa2->type)
-	return lsa1->type - lsa2->type;
-  
-      if (lsa1->id != lsa2->id)
-	return lsa1->id - lsa2->id;
-  
-      return lsa1->age - lsa2->age;
-    }
+  return lsa1->sn - lsa2->sn;
 }
 
 static inline void
-show_lsa_router(struct top_hash_entry *he)
+show_lsa_distance(struct top_hash_entry *he)
+{
+  if (he->color == INSPF)
+    cli_msg(-1016, "\t\tdistance %u", he->dist);
+  else
+    cli_msg(-1016, "\t\tunreachable");
+}
+
+static inline void
+show_lsa_router(struct proto_ospf *po, struct top_hash_entry *he, int first, int verbose)
 {
   struct ospf_lsa_header *lsa = &(he->lsa);
   struct ospf_lsa_rt *rt = he->lsa_body;
   struct ospf_lsa_rt_link *rr = (struct ospf_lsa_rt_link *) (rt + 1);
-  u32 i;
+  int max = lsa_rt_count(lsa);
+  int i;
 
-  for (i = 0; i < rt->links; i++)
+  if (first)
+  {
+    cli_msg(-1016, "");
+    cli_msg(-1016, "\trouter %R", he->lsa.rt);
+    show_lsa_distance(he);
+  }
+
+
+  for (i = 0; i < max; i++)
+    if (rr[i].type == LSART_VLNK)
+      cli_msg(-1016, "\t\tvlink %R metric %u", rr[i].id, rr[i].metric);
+
+  for (i = 0; i < max; i++)
     if (rr[i].type == LSART_PTP)
-      cli_msg(-1016, "\t\trouter %R metric %u ", rr[i].id, rr[i].metric);
+      cli_msg(-1016, "\t\trouter %R metric %u", rr[i].id, rr[i].metric);
 
-  for (i = 0; i < rt->links; i++)
+  for (i = 0; i < max; i++)
     if (rr[i].type == LSART_NET)
     {
-      struct proto_ospf *po = he->oa->po;
-      struct top_hash_entry *net_he = ospf_hash_find(po->gr, he->oa->areaid, rr[i].id, rr[i].id, LSA_T_NET);
+#ifdef OSPFv2
+      struct top_hash_entry *net_he = ospf_hash_find_net(po->gr, he->domain, rr[i].id);
+
       if (net_he)
       {
 	struct ospf_lsa_header *net_lsa = &(net_he->lsa);
 	struct ospf_lsa_net *net_ln = net_he->lsa_body;
-	cli_msg(-1016, "\t\tnetwork %I/%d metric %u ", ipa_and(ipa_from_u32(net_lsa->id), net_ln->netmask), ipa_mklen(net_ln->netmask), rr[i].metric);
+
+	cli_msg(-1016, "\t\tnetwork %I/%d metric %u", 
+		ipa_and(ipa_from_u32(net_lsa->id), net_ln->netmask),
+		ipa_mklen(net_ln->netmask), rr[i].metric);
       }
       else
-	cli_msg(-1016, "\t\tnetwork ??? metric %u ", rr[i].metric);
+	cli_msg(-1016, "\t\tnetwork [%R] metric %u", rr[i].id, rr[i].metric);
+
+#else /* OSPFv3 */
+      cli_msg(-1016, "\t\tnetwork [%R-%u] metric %u", rr[i].id, rr[i].nif, rr[i].metric);
+#endif
     }
 
-  for (i = 0; i < rt->links; i++)
-    if (rr[i].type == LSART_STUB)
-      cli_msg(-1016, "\t\tstubnet %I/%d metric %u ", ipa_from_u32(rr[i].id), ipa_mklen(ipa_from_u32(rr[i].data)), rr[i].metric);
+#ifdef OSPFv2
+  if (!verbose)
+    return;
 
-  for (i = 0; i < rt->links; i++)
-    if (rr[i].type == LSART_VLNK)
-      cli_msg(-1016, "\t\tvlink %I metric %u ", ipa_from_u32(rr[i].id), rr[i].metric);
+  for (i = 0; i < max; i++)
+    if (rr[i].type == LSART_STUB)
+      cli_msg(-1016, "\t\tstubnet %I/%d metric %u", ipa_from_u32(rr[i].id),
+	      ipa_mklen(ipa_from_u32(rr[i].data)), rr[i].metric);
+#endif
 }
 
 static inline void
@@ -1118,64 +1204,152 @@ show_lsa_network(struct top_hash_entry *he)
 {
   struct ospf_lsa_header *lsa = &(he->lsa);
   struct ospf_lsa_net *ln = he->lsa_body;
-  u32 *rts = (u32 *) (ln + 1);
-  u32 max = (lsa->length - sizeof(struct ospf_lsa_header) - sizeof(struct ospf_lsa_net)) / sizeof(u32);
   u32 i;
 
+#ifdef OSPFv2
   cli_msg(-1016, "");
   cli_msg(-1016, "\tnetwork %I/%d", ipa_and(ipa_from_u32(lsa->id), ln->netmask), ipa_mklen(ln->netmask));
   cli_msg(-1016, "\t\tdr %R", lsa->rt);
+#else /* OSPFv3 */
+  cli_msg(-1016, "");
+  cli_msg(-1016, "\tnetwork [%R-%u]", lsa->rt, lsa->id);
+#endif
 
-  for (i = 0; i < max; i++)
-    cli_msg(-1016, "\t\trouter %R", rts[i]);
+  show_lsa_distance(he);
+
+  for (i = 0; i < lsa_net_count(lsa); i++)
+    cli_msg(-1016, "\t\trouter %R", ln->routers[i]);
 }
 
 static inline void
 show_lsa_sum_net(struct top_hash_entry *he)
 {
-  struct ospf_lsa_header *lsa = &(he->lsa);
-  struct ospf_lsa_sum *sm = he->lsa_body;
+  ip_addr ip;
+  int pxlen;
 
-  cli_msg(-1016, "\t\txnetwork %I/%d", ipa_and(ipa_from_u32(lsa->id), sm->netmask), ipa_mklen(sm->netmask));
+#ifdef OSPFv2
+  struct ospf_lsa_sum *ls = he->lsa_body;
+  pxlen = ipa_mklen(ls->netmask);
+  ip = ipa_and(ipa_from_u32(he->lsa.id), ls->netmask);
+#else /* OSPFv3 */
+  u8 pxopts;
+  u16 rest;
+  struct ospf_lsa_sum_net *ls = he->lsa_body;
+  lsa_get_ipv6_prefix(ls->prefix, &ip, &pxlen, &pxopts, &rest);
+#endif
+
+  cli_msg(-1016, "\t\txnetwork %I/%d metric %u", ip, pxlen, ls->metric);
 }
 
 static inline void
 show_lsa_sum_rt(struct top_hash_entry *he)
 {
-  cli_msg(-1016, "\t\txrouter %R", he->lsa.id);
+  u32 dst_rid, options;
+
+#ifdef OSPFv2
+  struct ospf_lsa_sum *ls = he->lsa_body;
+  dst_rid = he->lsa.id;
+  options = 0;
+#else /* OSPFv3 */
+  struct ospf_lsa_sum_rt *ls = he->lsa_body;
+  dst_rid = ls->drid; 
+  options = ls->options & OPTIONS_MASK;
+#endif
+
+  cli_msg(-1016, "\t\txrouter %R metric %u", dst_rid, ls->metric);
 }
 
 
 static inline void
 show_lsa_external(struct top_hash_entry *he)
 {
-  struct ospf_lsa_header *lsa = &(he->lsa);
   struct ospf_lsa_ext *ext = he->lsa_body;
-  struct ospf_lsa_ext_tos *et = (struct ospf_lsa_ext_tos *) (ext + 1);
-
   char str_via[STD_ADDRESS_P_LENGTH + 8] = "";
   char str_tag[16] = "";
+  ip_addr ip, rt_fwaddr;
+  int pxlen, ebit, rt_fwaddr_valid;
+  u32 rt_tag, rt_metric;
+
+  he->domain = 0; /* Unmark the LSA */
+
+  rt_metric = ext->metric & METRIC_MASK;
+  ebit = ext->metric & LSA_EXT_EBIT;
+#ifdef OSPFv2
+  ip = ipa_and(ipa_from_u32(he->lsa.id), ext->netmask);
+  pxlen = ipa_mklen(ext->netmask);
+  rt_fwaddr = ext->fwaddr;
+  rt_fwaddr_valid = !ipa_equal(rt_fwaddr, IPA_NONE);
+  rt_tag = ext->tag;
+#else /* OSPFv3 */
+  u8 pxopts;
+  u16 rest;
+  u32 *buf = ext->rest;
+  buf = lsa_get_ipv6_prefix(buf, &ip, &pxlen, &pxopts, &rest);
+
+  rt_fwaddr_valid = ext->metric & LSA_EXT_FBIT;
+  if (rt_fwaddr_valid)
+    buf = lsa_get_ipv6_addr(buf, &rt_fwaddr);
+  else 
+    rt_fwaddr = IPA_NONE;
+
+  if (ext->metric & LSA_EXT_TBIT)
+    rt_tag = *buf++;
+  else
+    rt_tag = 0;
+#endif
   
-  if (ipa_nonzero(et->fwaddr))
-    bsprintf(str_via, " via %I", et->fwaddr);
+  if (rt_fwaddr_valid)
+    bsprintf(str_via, " via %I", rt_fwaddr);
 
-  if (et->tag)
-    bsprintf(str_tag, " tag %08x", et->tag);
+  if (rt_tag)
+    bsprintf(str_tag, " tag %08x", rt_tag);
 
-  cli_msg(-1016, "\t\texternal %I/%d metric%s %u%s%s",
-	  ipa_and(ipa_from_u32(lsa->id), ext->netmask),
-	  ipa_mklen(ext->netmask), et->etm.etos.ebit ? "2" : "",
-	  et->etm.metric & METRIC_MASK, str_via, str_tag);
+  cli_msg(-1016, "\t\texternal %I/%d metric%s %u%s%s", ip, pxlen,
+	  ebit ? "2" : "", rt_metric, str_via, str_tag);
 }
 
+#ifdef OSPFv3
+static inline void
+show_lsa_prefix(struct top_hash_entry *he, struct ospf_lsa_header *cnode)
+{
+  struct ospf_lsa_prefix *px = he->lsa_body;
+  ip_addr pxa;
+  int pxlen;
+  u8 pxopts;
+  u16 metric;
+  u32 *buf;
+  int i;
+
+  /* We check whether given prefix-LSA is related to the current node */
+  if ((px->ref_type != cnode->type) || (px->ref_rt != cnode->rt))
+    return;
+
+  if ((px->ref_type == LSA_T_RT) && (px->ref_id != 0))
+    return;
+
+  if ((px->ref_type == LSA_T_NET) && (px->ref_id != cnode->id))
+    return;
+
+  buf = px->rest;
+  for (i = 0; i < px->pxcount; i++)
+    {
+      buf = lsa_get_ipv6_prefix(buf, &pxa, &pxlen, &pxopts, &metric);
+
+      if (px->ref_type == LSA_T_RT)
+	cli_msg(-1016, "\t\tstubnet %I/%d metric %u", pxa, pxlen, metric);
+      else
+	cli_msg(-1016, "\t\taddress %I/%d", pxa, pxlen);
+    }
+}
+#endif
 
 void
-ospf_sh_state(struct proto *p, int verbose)
+ospf_sh_state(struct proto *p, int verbose, int reachable)
 {
   struct proto_ospf *po = (struct proto_ospf *) p;
-  struct top_graph *f = po->gr;
-  unsigned int i, j;
-  u32 last_rt = 0xFFFFFFFF;
+  struct ospf_lsa_header *cnode = NULL;
+  int num = po->gr->hash_entries;
+  unsigned int i, ix, j1, j2, jx;
   u32 last_area = 0xFFFFFFFF;
 
   if (p->proto_state != PS_UP)
@@ -1185,75 +1359,291 @@ ospf_sh_state(struct proto *p, int verbose)
     return;
   }
 
-  struct top_hash_entry *hea[f->hash_entries];
+  /* We store interesting area-scoped LSAs in array hea and 
+     global-scoped (LSA_T_EXT) LSAs in array hex */
+
+  struct top_hash_entry *hea[num];
+  struct top_hash_entry *hex[verbose ? num : 0];
   struct top_hash_entry *he;
 
-  j = 0;
-  for (i = 0; i < f->hash_size; i++)
-    for (he = f->hash_table[i]; he != NULL; he = he->next)
-      hea[j++] = he;
-
-  if (j == f->hash_size)
-    die("Fatal mismatch");
-
-  qsort(hea, j, sizeof(struct top_hash_entry *), he_compare);
-
-  for (i = 0; i < j; i++)
+  j1 = j2 = jx = 0;
+  WALK_SLIST(he, po->lsal)
   {
-    if ((verbose == 0) && (hea[i]->lsa.type > LSA_T_NET))
-      continue;
+    int accept;
 
-    if (last_area != hea[i]->oa->areaid)
-    {
-      cli_msg(-1016, "");
-      cli_msg(-1016, "area %R", hea[i]->oa->areaid);
-      last_area = hea[i]->oa->areaid;
-      last_rt = 0xFFFFFFFF;
-    }
-
-    if ((hea[i]->lsa.rt != last_rt) && (hea[i]->lsa.type != LSA_T_NET))
-    {
-      cli_msg(-1016, "");
-      cli_msg(-1016, (hea[i]->lsa.type != LSA_T_EXT) ? "\trouter %R" : "\txrouter %R", hea[i]->lsa.rt);
-      last_rt = hea[i]->lsa.rt;
-    }
-
-    switch (hea[i]->lsa.type)
-    {
+    switch (he->lsa.type)
+      {
       case LSA_T_RT:
-	show_lsa_router(hea[i]);
-	break;
-
       case LSA_T_NET:
-	show_lsa_network(hea[i]);
+	accept = 1;
 	break;
 
       case LSA_T_SUM_NET:
-	show_lsa_sum_net(hea[i]);
+      case LSA_T_SUM_RT:
+#ifdef OSPFv3
+      case LSA_T_PREFIX:
+#endif
+	accept = verbose;
+	break;
+
+      case LSA_T_EXT:
+	if (verbose)
+	{
+	  he->domain = 1; /* Abuse domain field to mark the LSA */
+	  hex[jx++] = he;
+	}
+      default:
+	accept = 0;
+      }
+
+    if (accept)
+      hea[j1++] = he;
+    else
+      j2++;
+  }
+
+  if ((j1 + j2) != num)
+    die("Fatal mismatch");
+
+  qsort(hea, j1, sizeof(struct top_hash_entry *), lsa_compare_for_state);
+  qsort(hex, jx, sizeof(struct top_hash_entry *), ext_compare_for_state);
+
+  /*
+   * This code is a bit tricky, we have a primary LSAs (router and
+   * network) that are presented as a node, and secondary LSAs that
+   * are presented as a part of a primary node. cnode represents an
+   * currently opened node (whose header was presented). The LSAs are
+   * sorted to get secondary LSAs just after related primary LSA (if
+   * available). We present secondary LSAs only when related primary
+   * LSA is opened.
+   *
+   * AS-external LSAs are stored separately as they might be presented
+   * several times (for each area when related ASBR is opened). When
+   * the node is closed, related external routes are presented. We
+   * also have to take into account that in OSPFv3, there might be
+   * more router-LSAs and only the first should be considered as a
+   * primary. This is handled by not closing old router-LSA when next
+   * one is processed (which is not opened because there is already
+   * one opened).
+   */
+
+  ix = 0;
+  for (i = 0; i < j1; i++)
+  {
+    he = hea[i];
+
+    /* If there is no opened node, we open the LSA (if appropriate) or skip to the next one */
+    if (!cnode)
+    {
+      if (((he->lsa.type == LSA_T_RT) || (he->lsa.type == LSA_T_NET))
+	  && ((he->color == INSPF) || !reachable))
+      {
+	cnode = &(he->lsa);
+
+	if (he->domain != last_area)
+	{
+	  cli_msg(-1016, "");
+	  cli_msg(-1016, "area %R", he->domain);
+	  last_area = he->domain;
+	  ix = 0;
+	}
+      }
+      else
+	continue;
+    }
+
+    ASSERT(cnode && (he->domain == last_area) && (he->lsa.rt == cnode->rt));
+
+    switch (he->lsa.type)
+    {
+      case LSA_T_RT:
+	show_lsa_router(po, he, he->lsa.id == cnode->id, verbose);
+	break;
+
+      case LSA_T_NET:
+	show_lsa_network(he);
+	break;
+
+      case LSA_T_SUM_NET:
+	if (cnode->type == LSA_T_RT)
+	  show_lsa_sum_net(he);
 	break;
 
       case LSA_T_SUM_RT:
-	show_lsa_sum_rt(hea[i]);
+	if (cnode->type == LSA_T_RT)
+	  show_lsa_sum_rt(he);
 	break;
+
+#ifdef OSPFv3
+      case LSA_T_PREFIX:
+	show_lsa_prefix(he, cnode);
+	break;
+#endif
+
       case LSA_T_EXT:
-	show_lsa_external(hea[i]);
+	show_lsa_external(he);
 	break;
     }
+
+    /* In these cases, we close the current node */
+    if ((i+1 == j1)
+	|| (hea[i+1]->domain != last_area)
+	|| (hea[i+1]->lsa.rt != cnode->rt)
+	|| (hea[i+1]->lsa.type == LSA_T_NET))
+    {
+      while ((ix < jx) && (hex[ix]->lsa.rt < cnode->rt))
+	ix++;
+
+      while ((ix < jx) && (hex[ix]->lsa.rt == cnode->rt))
+	show_lsa_external(hex[ix++]);
+
+      cnode = NULL;
+    }
+  }
+
+  int hdr = 0;
+  u32 last_rt = 0xFFFFFFFF;
+  for (ix = 0; ix < jx; ix++)
+  {
+    he = hex[ix];
+
+    /* If it is still marked, we show it now. */
+    if (he->domain)
+    {
+      he->domain = 0;
+
+      if ((he->color != INSPF) && reachable)
+	continue;
+
+      if (!hdr)
+      {
+	cli_msg(-1016, "");
+	cli_msg(-1016, "other ASBRs");
+	hdr = 1;
+      }
+
+      if (he->lsa.rt != last_rt)
+      {
+	cli_msg(-1016, "");
+	cli_msg(-1016, "\trouter %R", he->lsa.rt);
+	last_rt = he->lsa.rt;
+      }
+
+      show_lsa_external(he);
+    }
   }
 
   cli_msg(0, "");
 }
 
+
+static int
+lsa_compare_for_lsadb(const void *p1, const void *p2)
+{
+  struct top_hash_entry * he1 = * (struct top_hash_entry **) p1;
+  struct top_hash_entry * he2 = * (struct top_hash_entry **) p2;
+  struct ospf_lsa_header *lsa1 = &(he1->lsa);
+  struct ospf_lsa_header *lsa2 = &(he2->lsa);
+  int sc1 = LSA_SCOPE(lsa1);
+  int sc2 = LSA_SCOPE(lsa2);
+
+  if (sc1 != sc2)
+    return sc2 - sc1;
+
+  if (he1->domain != he2->domain)
+    return he1->domain - he2->domain;
+
+  if (lsa1->rt != lsa2->rt)
+    return lsa1->rt - lsa2->rt;
+  
+  if (lsa1->id != lsa2->id)
+    return lsa1->id - lsa2->id;
+
+  if (lsa1->type != lsa2->type)
+    return lsa1->type - lsa2->type;
+
+  return lsa1->sn - lsa2->sn;
+}
+
+void
+ospf_sh_lsadb(struct proto *p)
+{
+  struct proto_ospf *po = (struct proto_ospf *) p;
+  int num = po->gr->hash_entries;
+  unsigned int i, j;
+  int last_dscope = -1;
+  u32 last_domain = 0;
+
+  if (p->proto_state != PS_UP)
+  {
+    cli_msg(-1017, "%s: is not up", p->name);
+    cli_msg(0, "");
+    return;
+  }
+
+  struct top_hash_entry *hea[num];
+  struct top_hash_entry *he;
+
+  j = 0;
+  WALK_SLIST(he, po->lsal)
+    hea[j++] = he;
+
+  if (j != num)
+    die("Fatal mismatch");
+
+  qsort(hea, j, sizeof(struct top_hash_entry *), lsa_compare_for_lsadb);
+
+  for (i = 0; i < j; i++)
+  {
+    struct ospf_lsa_header *lsa = &(hea[i]->lsa);
+    int dscope = LSA_SCOPE(lsa);
+    
+    if ((dscope != last_dscope) || (hea[i]->domain != last_domain))
+    {
+      cli_msg(-1017, "");
+      switch (dscope)
+      {
+	case LSA_SCOPE_AS:
+	  cli_msg(-1017, "Global");
+	  break;
+	case LSA_SCOPE_AREA:
+	  cli_msg(-1017, "Area %R", hea[i]->domain);
+	  break;
+#ifdef OSPFv3
+	case LSA_SCOPE_LINK:
+	  {
+	    struct iface *ifa = if_find_by_index(hea[i]->domain);
+	    cli_msg(-1017, "Link %s", (ifa != NULL) ? ifa->name : "?");
+	  }
+	  break;
+#endif
+      }
+      cli_msg(-1017, "");
+      cli_msg(-1017," Type   LS ID           Router           Age  Sequence  Checksum");
+
+      last_dscope = dscope;
+      last_domain = hea[i]->domain;
+    }
+
+
+    cli_msg(-1017," %04x  %-15R %-15R %5u  %08x    %04x",
+	    lsa->type, lsa->id, lsa->rt, lsa->age, lsa->sn, lsa->checksum);
+  }
+  cli_msg(0, "");
+}
+
+
 struct protocol proto_ospf = {
-  name:"OSPF",
-  template:"ospf%d",
-  attr_class:EAP_OSPF,
-  init:ospf_init,
-  dump:ospf_dump,
-  start:ospf_start,
-  shutdown:ospf_shutdown,
-  get_route_info:ospf_get_route_info,
-  get_attr:ospf_get_attr,
-  get_status:ospf_get_status,
-  reconfigure:ospf_reconfigure
+  name:			"OSPF",
+  template:		"ospf%d",
+  attr_class:		EAP_OSPF,
+  init:			ospf_init,
+  dump:			ospf_dump,
+  start:		ospf_start,
+  shutdown:		ospf_shutdown,
+  reconfigure:		ospf_reconfigure,
+  get_status:		ospf_get_status,
+  get_attr:		ospf_get_attr,
+  get_route_info:	ospf_get_route_info
+  // show_proto_info:	ospf_sh
 };
