@@ -7,6 +7,10 @@
  *	Can be freely distributed and used under the terms of the GNU GPL.
  */
 
+/* Unfortunately, some glibc versions hide parts of RFC 3542 API
+   if _GNU_SOURCE is not defined. */
+#define _GNU_SOURCE 1
+
 #include <stdio.h>
 #include <stdlib.h>
 #include <time.h>
@@ -14,6 +18,7 @@
 #include <sys/types.h>
 #include <sys/socket.h>
 #include <sys/fcntl.h>
+#include <sys/uio.h>
 #include <sys/un.h>
 #include <unistd.h>
 #include <errno.h>
@@ -70,7 +75,8 @@ static struct resclass rf_class = {
   "FILE",
   sizeof(struct rfile),
   rf_free,
-  rf_dump
+  rf_dump,
+  NULL
 };
 
 void *
@@ -111,7 +117,8 @@ tracked_fopen(pool *p, char *name, char *mode)
 static list near_timers, far_timers;
 static bird_clock_t first_far_timer = TIME_INFINITY;
 
-bird_clock_t now, now_real;
+/* now must be different from 0, because 0 is a special value in timer->expires */
+bird_clock_t now = 1, now_real;
 
 static void
 update_times_plain(void)
@@ -195,7 +202,8 @@ static struct resclass tm_class = {
   "Timer",
   sizeof(timer),
   tm_free,
-  tm_dump
+  tm_dump,
+  NULL
 };
 
 /**
@@ -410,23 +418,22 @@ tm_parse_date(char *x)
   return t;
 }
 
-/**
- * tm_format_date - convert date to textual representation
- * @x: destination buffer of size %TM_DATE_BUFFER_SIZE
- * @t: time
- *
- * This function formats the given relative time value @t to a textual
- * date representation (dd-mm-yyyy) in real time..
- */
-void
-tm_format_date(char *x, bird_clock_t t)
+static void
+tm_format_reltime(char *x, struct tm *tm, bird_clock_t delta)
 {
-  struct tm *tm;
+  static char *month_names[12] = { "Jan", "Feb", "Mar", "Apr", "May", "Jun",
+				   "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };
 
-  tm = localtime(&t);
-  bsprintf(x, "%02d-%02d-%04d", tm->tm_mday, tm->tm_mon+1, tm->tm_year+1900);
+  if (delta < 20*3600)
+    bsprintf(x, "%02d:%02d", tm->tm_hour, tm->tm_min);
+  else if (delta < 360*86400)
+    bsprintf(x, "%s%02d", month_names[tm->tm_mon], tm->tm_mday);
+  else
+    bsprintf(x, "%d", tm->tm_year+1900);
 }
 
+#include "conf/conf.h"
+
 /**
  * tm_format_datetime - convert date and time to textual representation
  * @x: destination buffer of size %TM_DATETIME_BUFFER_SIZE
@@ -436,39 +443,25 @@ tm_format_date(char *x, bird_clock_t t)
  * date/time representation (dd-mm-yyyy hh:mm:ss) in real time.
  */
 void
-tm_format_datetime(char *x, bird_clock_t t)
+tm_format_datetime(char *x, struct timeformat *fmt_spec, bird_clock_t t)
 {
+  const char *fmt_used;
   struct tm *tm;
   bird_clock_t delta = now - t;
   t = now_real - delta;
   tm = localtime(&t);
-  if (strftime(x, TM_DATETIME_BUFFER_SIZE, "%d-%m-%Y %H:%M:%S", tm) == TM_DATETIME_BUFFER_SIZE)
-    strcpy(x, "<too-long>");
-}
 
-/**
- * tm_format_reltime - convert date and time to relative textual representation
- * @x: destination buffer of size %TM_RELTIME_BUFFER_SIZE
- * @t: time
- *
- * This function formats the given relative time value @t to a short
- * textual representation in real time, relative to the current time.
- */
-void
-tm_format_reltime(char *x, bird_clock_t t)
-{
-  struct tm *tm;
-  static char *month_names[12] = { "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };
+  if (fmt_spec->fmt1 == NULL)
+    return tm_format_reltime(x, tm, delta);
 
-  bird_clock_t delta = now - t;
-  t = now_real - delta;
-  tm = localtime(&t);
-  if (delta < 20*3600)
-    bsprintf(x, "%02d:%02d", tm->tm_hour, tm->tm_min);
-  else if (delta < 360*86400)
-    bsprintf(x, "%s%02d", month_names[tm->tm_mon], tm->tm_mday);
+  if ((fmt_spec->limit == 0) || (delta < fmt_spec->limit))
+    fmt_used = fmt_spec->fmt1;
   else
-    bsprintf(x, "%d", tm->tm_year+1900);
+    fmt_used = fmt_spec->fmt2;
+
+  int rv = strftime(x, TM_DATETIME_BUFFER_SIZE, fmt_used, tm);
+  if (((rv == 0) && fmt_used[0]) || (rv == TM_DATETIME_BUFFER_SIZE))
+    strcpy(x, "<too-long>");
 }
 
 /**
@@ -492,10 +485,6 @@ tm_format_reltime(char *x, bird_clock_t t)
 #define SOL_IPV6 IPPROTO_IPV6
 #endif
 
-#ifndef IPV6_ADD_MEMBERSHIP
-#define IPV6_ADD_MEMBERSHIP IP_ADD_MEMBERSHIP
-#endif
-
 static list sock_list;
 static struct birdsock *current_sock;
 static struct birdsock *stored_sock;
@@ -583,7 +572,8 @@ static struct resclass sk_class = {
   "Socket",
   sizeof(sock),
   sk_free,
-  sk_dump
+  sk_dump,
+  NULL
 };
 
 /**
@@ -627,6 +617,12 @@ fill_in_sockaddr(sockaddr *sa, ip_addr a, unsigned port)
   set_inaddr(&sa->sin6_addr, a);
 }
 
+static inline void
+fill_in_sockifa(sockaddr *sa, struct iface *ifa)
+{
+  sa->sin6_scope_id = ifa ? ifa->index : 0;
+}
+
 void
 get_sockaddr(struct sockaddr_in6 *sa, ip_addr *a, unsigned *port, int check)
 {
@@ -652,6 +648,11 @@ fill_in_sockaddr(sockaddr *sa, ip_addr a, unsigned port)
   set_inaddr(&sa->sin_addr, a);
 }
 
+static inline void
+fill_in_sockifa(sockaddr *sa UNUSED, struct iface *ifa UNUSED)
+{
+}
+
 void
 get_sockaddr(struct sockaddr_in *sa, ip_addr *a, unsigned *port, int check)
 {
@@ -665,18 +666,100 @@ get_sockaddr(struct sockaddr_in *sa, ip_addr *a, unsigned *port, int check)
 
 #endif
 
+
+#ifdef IPV6
+
+/* PKTINFO handling is also standardized in IPv6 */
+#define CMSG_RX_SPACE CMSG_SPACE(sizeof(struct in6_pktinfo))
+#define CMSG_TX_SPACE CMSG_SPACE(sizeof(struct in6_pktinfo))
+
+/*
+ * RFC 2292 uses IPV6_PKTINFO for both the socket option and the cmsg
+ * type, RFC 3542 changed the socket option to IPV6_RECVPKTINFO. If we
+ * don't have IPV6_RECVPKTINFO we suppose the OS implements the older
+ * RFC and we use IPV6_PKTINFO.
+ */
+#ifndef IPV6_RECVPKTINFO
+#define IPV6_RECVPKTINFO IPV6_PKTINFO
+#endif
+
+static char *
+sysio_register_cmsgs(sock *s)
+{
+  int ok = 1;
+  if ((s->flags & SKF_LADDR_RX) &&
+      setsockopt(s->fd, IPPROTO_IPV6, IPV6_RECVPKTINFO, &ok, sizeof(ok)) < 0)
+    return "IPV6_RECVPKTINFO";
+
+  return NULL;
+}
+
+static void
+sysio_process_rx_cmsgs(sock *s, struct msghdr *msg)
+{
+  struct cmsghdr *cm;
+  struct in6_pktinfo *pi = NULL;
+
+  if (!(s->flags & SKF_LADDR_RX))
+    return;
+
+  for (cm = CMSG_FIRSTHDR(msg); cm != NULL; cm = CMSG_NXTHDR(msg, cm))
+    {
+      if (cm->cmsg_level == IPPROTO_IPV6 && cm->cmsg_type == IPV6_PKTINFO)
+	pi = (struct in6_pktinfo *) CMSG_DATA(cm);
+    }
+
+  if (!pi)
+    {
+      s->laddr = IPA_NONE;
+      s->lifindex = 0;
+      return;
+    }
+
+  get_inaddr(&s->laddr, &pi->ipi6_addr);
+  s->lifindex = pi->ipi6_ifindex;
+  return;
+}
+
+/*
+static void
+sysio_prepare_tx_cmsgs(sock *s, struct msghdr *msg, void *cbuf, size_t cbuflen)
+{
+  struct cmsghdr *cm;
+  struct in6_pktinfo *pi;
+
+  if (!(s->flags & SKF_LADDR_TX))
+    return;
+
+  msg->msg_control = cbuf;
+  msg->msg_controllen = cbuflen;
+
+  cm = CMSG_FIRSTHDR(msg);
+  cm->cmsg_level = IPPROTO_IPV6;
+  cm->cmsg_type = IPV6_PKTINFO;
+  cm->cmsg_len = CMSG_LEN(sizeof(*pi));
+
+  pi = (struct in6_pktinfo *) CMSG_DATA(cm);
+  set_inaddr(&pi->ipi6_addr, s->saddr);
+  pi->ipi6_ifindex = s->iface ? s->iface->index : 0;
+
+  msg->msg_controllen = cm->cmsg_len;
+  return;
+}
+*/
+#endif
+
 static char *
 sk_set_ttl_int(sock *s)
 {
-  int one = 1;
 #ifdef IPV6
-  if (s->type != SK_UDP_MC && s->type != SK_IP_MC &&
-      setsockopt(s->fd, SOL_IPV6, IPV6_UNICAST_HOPS, &s->ttl, sizeof(s->ttl)) < 0)
+  if (setsockopt(s->fd, SOL_IPV6, IPV6_UNICAST_HOPS, &s->ttl, sizeof(s->ttl)) < 0)
     return "IPV6_UNICAST_HOPS";
 #else
   if (setsockopt(s->fd, SOL_IP, IP_TTL, &s->ttl, sizeof(s->ttl)) < 0)
     return "IP_TTL";
 #ifdef CONFIG_UNIX_DONTROUTE
+  int one = 1;
   if (s->ttl == 1 && setsockopt(s->fd, SOL_SOCKET, SO_DONTROUTE, &one, sizeof(one)) < 0)
     return "SO_DONTROUTE";
 #endif 
@@ -691,7 +774,7 @@ static char *
 sk_setup(sock *s)
 {
   int fd = s->fd;
-  char *err;
+  char *err = NULL;
 
   if (fcntl(fd, F_SETFL, O_NONBLOCK) < 0)
     ERR("fcntl(O_NONBLOCK)");
@@ -710,9 +793,8 @@ sk_setup(sock *s)
 
   if (s->ttl >= 0)
     err = sk_set_ttl_int(s);
-  else
-    err = NULL;
 
+  sysio_register_cmsgs(s);
 bad:
   return err;
 }
@@ -768,6 +850,153 @@ sk_set_md5_auth(sock *s, ip_addr a, char *passwd)
   return sk_set_md5_auth_int(s, &sa, passwd);
 }
 
+int
+sk_set_broadcast(sock *s, int enable)
+{
+  if (setsockopt(s->fd, SOL_SOCKET, SO_BROADCAST, &enable, sizeof(enable)) < 0)
+    {
+      log(L_ERR "sk_set_broadcast: SO_BROADCAST: %m");
+      return -1;
+    }
+
+  return 0;
+}
+
+
+#ifdef IPV6
+
+int
+sk_set_ipv6_checksum(sock *s, int offset)
+{
+  if (setsockopt(s->fd, IPPROTO_IPV6, IPV6_CHECKSUM, &offset, sizeof(offset)) < 0)
+    {
+      log(L_ERR "sk_set_ipv6_checksum: IPV6_CHECKSUM: %m");
+      return -1;
+    }
+
+  return 0;
+}
+
+int
+sk_setup_multicast(sock *s)
+{
+  char *err;
+  int zero = 0;
+  int index;
+
+  ASSERT(s->iface && s->iface->addr);
+
+  index = s->iface->index;
+  if (setsockopt(s->fd, SOL_IPV6, IPV6_MULTICAST_HOPS, &s->ttl, sizeof(s->ttl)) < 0)
+    ERR("IPV6_MULTICAST_HOPS");
+  if (setsockopt(s->fd, SOL_IPV6, IPV6_MULTICAST_LOOP, &zero, sizeof(zero)) < 0)
+    ERR("IPV6_MULTICAST_LOOP");
+  if (setsockopt(s->fd, SOL_IPV6, IPV6_MULTICAST_IF, &index, sizeof(index)) < 0)
+    ERR("IPV6_MULTICAST_IF");
+
+  if (err = sysio_bind_to_iface(s))
+    goto bad;
+
+  return 0;
+
+bad:
+  log(L_ERR "sk_setup_multicast: %s: %m", err);
+  return -1;
+}
+
+int
+sk_join_group(sock *s, ip_addr maddr)
+{
+  struct ipv6_mreq mreq;
+	
+  set_inaddr(&mreq.ipv6mr_multiaddr, maddr);
+
+#ifdef CONFIG_IPV6_GLIBC_20
+  mreq.ipv6mr_ifindex = s->iface->index;
+#else
+  mreq.ipv6mr_interface = s->iface->index;
+#endif
+
+  if (setsockopt(s->fd, SOL_IPV6, IPV6_JOIN_GROUP, &mreq, sizeof(mreq)) < 0)
+    {
+      log(L_ERR "sk_join_group: IPV6_JOIN_GROUP: %m");
+      return -1;
+    }
+
+  return 0;
+}
+
+int
+sk_leave_group(sock *s, ip_addr maddr)
+{
+  struct ipv6_mreq mreq;
+	
+  set_inaddr(&mreq.ipv6mr_multiaddr, maddr);
+
+#ifdef CONFIG_IPV6_GLIBC_20
+  mreq.ipv6mr_ifindex = s->iface->index;
+#else
+  mreq.ipv6mr_interface = s->iface->index;
+#endif
+
+  if (setsockopt(s->fd, SOL_IPV6, IPV6_LEAVE_GROUP, &mreq, sizeof(mreq)) < 0)
+    {
+      log(L_ERR "sk_leave_group: IPV6_LEAVE_GROUP: %m");
+      return -1;
+    }
+
+  return 0;
+}
+
+
+#else /* IPV4 */
+
+int
+sk_setup_multicast(sock *s)
+{
+  char *err;
+
+  ASSERT(s->iface && s->iface->addr);
+
+  if (err = sysio_setup_multicast(s))
+    {
+      log(L_ERR "sk_setup_multicast: %s: %m", err);
+      return -1;
+    }
+
+  return 0;
+}
+
+int
+sk_join_group(sock *s, ip_addr maddr)
+{
+ char *err;
+
+ if (err = sysio_join_group(s, maddr))
+    {
+      log(L_ERR "sk_join_group: %s: %m", err);
+      return -1;
+    }
+
+  return 0;
+}
+
+int
+sk_leave_group(sock *s, ip_addr maddr)
+{
+ char *err;
+
+ if (err = sysio_leave_group(s, maddr))
+    {
+      log(L_ERR "sk_leave_group: %s: %m", err);
+      return -1;
+    }
+
+  return 0;
+}
+
+#endif 
+
 
 static void
 sk_tcp_connected(sock *s)
@@ -792,7 +1021,14 @@ sk_passive_connected(sock *s, struct sockaddr *sa, int al, int type)
       t->rbsize = s->rbsize;
       t->tbsize = s->tbsize;
       if (type == SK_TCP)
-	get_sockaddr((sockaddr *) sa, &t->daddr, &t->dport, 1);
+	{
+	  sockaddr lsa;
+	  int lsa_len = sizeof(lsa);
+	  if (getsockname(fd, (struct sockaddr *) &lsa, &lsa_len) == 0)
+	    get_sockaddr(&lsa, &t->saddr, &t->sport, 1);
+
+	  get_sockaddr((sockaddr *) sa, &t->daddr, &t->dport, 1);
+	}
       sk_insert(t);
       if (err = sk_setup(t))
 	{
@@ -806,7 +1042,6 @@ sk_passive_connected(sock *s, struct sockaddr *sa, int al, int type)
     }
   else if (errno != EINTR && errno != EAGAIN)
     {
-      log(L_ERR "accept: %m");
       s->err_hook(s, errno);
     }
   return 0;
@@ -841,11 +1076,9 @@ sk_open(sock *s)
       fd = socket(BIRD_PF, SOCK_STREAM, IPPROTO_TCP);
       break;
     case SK_UDP:
-    case SK_UDP_MC:
       fd = socket(BIRD_PF, SOCK_DGRAM, IPPROTO_UDP);
       break;
     case SK_IP:
-    case SK_IP_MC:
       fd = socket(BIRD_PF, SOCK_RAW, s->dport);
       break;
     case SK_MAGIC:
@@ -861,61 +1094,11 @@ sk_open(sock *s)
   if (err = sk_setup(s))
     goto bad;
 
-  switch (type)
-    {
-    case SK_UDP:
-    case SK_IP:
-      if (s->iface)			/* It's a broadcast socket */
-#ifdef IPV6
-	bug("IPv6 has no broadcasts");
-#else
-	if (setsockopt(fd, SOL_SOCKET, SO_BROADCAST, &one, sizeof(one)) < 0)
-	  ERR("SO_BROADCAST");
-#endif
-      break;
-    case SK_UDP_MC:
-    case SK_IP_MC:
-      {
-#ifdef IPV6
-	/* Fortunately, IPv6 socket interface is recent enough and therefore standardized */
-	ASSERT(s->iface && s->iface->addr);
-	if (ipa_nonzero(s->daddr))
-	  {
-	    int t = s->iface->index;
-	    int zero = 0;
-	    if (setsockopt(fd, SOL_IPV6, IPV6_MULTICAST_HOPS, &s->ttl, sizeof(s->ttl)) < 0)
-	      ERR("IPV6_MULTICAST_HOPS");
-	    if (setsockopt(fd, SOL_IPV6, IPV6_MULTICAST_LOOP, &zero, sizeof(zero)) < 0)
-	      ERR("IPV6_MULTICAST_LOOP");
-	    if (setsockopt(fd, SOL_IPV6, IPV6_MULTICAST_IF, &t, sizeof(t)) < 0)
-	      ERR("IPV6_MULTICAST_IF");
-	  }
-	if (has_src)
-	  {
-	    struct ipv6_mreq mreq;
-	    set_inaddr(&mreq.ipv6mr_multiaddr, s->daddr);
-#ifdef CONFIG_IPV6_GLIBC_20
-	    mreq.ipv6mr_ifindex = s->iface->index;
-#else
-	    mreq.ipv6mr_interface = s->iface->index;
-#endif /* CONFIG_IPV6_GLIBC_20 */
-	    if (setsockopt(fd, SOL_IPV6, IPV6_ADD_MEMBERSHIP, &mreq, sizeof(mreq)) < 0)
-	      ERR("IPV6_ADD_MEMBERSHIP");
-	  }
-#else
-	/* With IPv4 there are zillions of different socket interface variants. Ugh. */
-	ASSERT(s->iface && s->iface->addr);
-	if (err = sysio_mcast_join(s))
-	  goto bad;
-#endif /* IPV6 */
-      break;
-      }
-    }
   if (has_src)
     {
       int port;
 
-      if (type == SK_IP || type == SK_IP_MC)
+      if (type == SK_IP)
 	port = 0;
       else
 	{
@@ -924,12 +1107,8 @@ sk_open(sock *s)
 	    ERR("SO_REUSEADDR");
 	}
       fill_in_sockaddr(&sa, s->saddr, port);
-#ifdef CONFIG_SKIP_MC_BIND
-      if ((type != SK_UDP_MC) && (type != SK_IP_MC) &&
-	  bind(fd, (struct sockaddr *) &sa, sizeof(sa)) < 0)
-#else
+      fill_in_sockifa(&sa, s->iface);
       if (bind(fd, (struct sockaddr *) &sa, sizeof(sa)) < 0)
-#endif
 	ERR("bind");
     }
   fill_in_sockaddr(&sa, s->daddr, s->dport);
@@ -988,7 +1167,7 @@ bad_no_log:
   return -1;
 }
 
-int
+void
 sk_open_unix(sock *s, char *name)
 {
   int fd;
@@ -997,15 +1176,13 @@ sk_open_unix(sock *s, char *name)
 
   fd = socket(AF_UNIX, SOCK_STREAM, 0);
   if (fd < 0)
-    die("sk_open_unix: socket: %m");
+    ERR("socket");
   s->fd = fd;
   if (err = sk_setup(s))
     goto bad;
   unlink(name);
- 
-  if (strlen(name) >= sizeof(sa.sun_path))
-    die("sk_open_unix: path too long");
 
+  /* Path length checked in test_old_bird() */
   sa.sun_family = AF_UNIX;
   strcpy(sa.sun_path, name);
   if (bind(fd, (struct sockaddr *) &sa, SUN_LEN(&sa)) < 0)
@@ -1013,15 +1190,15 @@ sk_open_unix(sock *s, char *name)
   if (listen(fd, 8))
     ERR("listen");
   sk_insert(s);
-  return 0;
+  return;
 
-bad:
+ bad:
   log(L_ERR "sk_open_unix: %s: %m", err);
-  close(fd);
-  s->fd = -1;
-  return -1;
+  die("Unable to create control socket %s", name);
 }
 
+static inline void reset_tx_buffer(sock *s) { s->ttx = s->tpos = s->tbuf; }
+
 static int
 sk_maybe_write(sock *s)
 {
@@ -1039,39 +1216,50 @@ sk_maybe_write(sock *s)
 	    {
 	      if (errno != EINTR && errno != EAGAIN)
 		{
-                  s->ttx = s->tpos;	/* empty tx buffer */
-		  s->err_hook(s, errno);
+		  reset_tx_buffer(s);
+		  /* EPIPE is just a connection close notification during TX */
+		  s->err_hook(s, (errno != EPIPE) ? errno : 0);
 		  return -1;
 		}
 	      return 0;
 	    }
 	  s->ttx += e;
 	}
-      s->ttx = s->tpos = s->tbuf;
+      reset_tx_buffer(s);
       return 1;
     case SK_UDP:
-    case SK_UDP_MC:
     case SK_IP:
-    case SK_IP_MC:
       {
-	sockaddr sa;
-
 	if (s->tbuf == s->tpos)
 	  return 1;
-	fill_in_sockaddr(&sa, s->faddr, s->fport);
 
-	e = sendto(s->fd, s->tbuf, s->tpos - s->tbuf, 0, (struct sockaddr *) &sa, sizeof(sa));
+	sockaddr sa;
+	fill_in_sockaddr(&sa, s->daddr, s->dport);
+	fill_in_sockifa(&sa, s->iface);
+
+	struct iovec iov = {s->tbuf, s->tpos - s->tbuf};
+	// byte cmsg_buf[CMSG_TX_SPACE];
+
+	struct msghdr msg = {
+	  .msg_name = &sa,
+	  .msg_namelen = sizeof(sa),
+	  .msg_iov = &iov,
+	  .msg_iovlen = 1};
+
+	// sysio_prepare_tx_cmsgs(s, &msg, cmsg_buf, sizeof(cmsg_buf));
+	e = sendmsg(s->fd, &msg, 0);
+
 	if (e < 0)
 	  {
 	    if (errno != EINTR && errno != EAGAIN)
 	      {
-                s->ttx = s->tpos;	/* empty tx buffer */
+		reset_tx_buffer(s);
 		s->err_hook(s, errno);
 		return -1;
 	      }
 	    return 0;
 	  }
-	s->tpos = s->tbuf;
+	reset_tx_buffer(s);
 	return 1;
       }
     default:
@@ -1117,8 +1305,6 @@ sk_rx_ready(sock *s)
 int
 sk_send(sock *s, unsigned len)
 {
-  s->faddr = s->daddr;
-  s->fport = s->dport;
   s->ttx = s->tbuf;
   s->tpos = s->tbuf + len;
   return sk_maybe_write(s);
@@ -1137,13 +1323,28 @@ sk_send(sock *s, unsigned len)
 int
 sk_send_to(sock *s, unsigned len, ip_addr addr, unsigned port)
 {
-  s->faddr = addr;
-  s->fport = port;
+  s->daddr = addr;
+  s->dport = port;
   s->ttx = s->tbuf;
   s->tpos = s->tbuf + len;
   return sk_maybe_write(s);
 }
 
+/*
+int
+sk_send_full(sock *s, unsigned len, struct iface *ifa,
+	     ip_addr saddr, ip_addr daddr, unsigned dport)
+{
+  s->iface = ifa;
+  s->saddr = saddr;
+  s->daddr = daddr;
+  s->dport = dport;
+  s->ttx = s->tbuf;
+  s->tpos = s->tbuf + len;
+  return sk_maybe_write(s);
+}
+*/
+
 static int
 sk_read(sock *s)
 {
@@ -1189,8 +1390,21 @@ sk_read(sock *s)
     default:
       {
 	sockaddr sa;
-	int al = sizeof(sa);
-	int e = recvfrom(s->fd, s->rbuf, s->rbsize, 0, (struct sockaddr *) &sa, &al);
+	int e;
+
+	struct iovec iov = {s->rbuf, s->rbsize};
+	byte cmsg_buf[CMSG_RX_SPACE];
+
+	struct msghdr msg = {
+	  .msg_name = &sa,
+	  .msg_namelen = sizeof(sa),
+	  .msg_iov = &iov,
+	  .msg_iovlen = 1,
+	  .msg_control = cmsg_buf,
+	  .msg_controllen = sizeof(cmsg_buf),
+	  .msg_flags = 0};
+
+	e = recvmsg(s->fd, &msg, 0);
 
 	if (e < 0)
 	  {
@@ -1200,6 +1414,8 @@ sk_read(sock *s)
 	  }
 	s->rpos = s->rbuf + e;
 	get_sockaddr(&sa, &s->faddr, &s->fport, 1);
+	sysio_process_rx_cmsgs(s, &msg);
+
 	s->rx_hook(s, e);
 	return 1;
       }
@@ -1409,7 +1625,6 @@ io_loop(void)
 	    {
 	      sock *s = current_sock;
 	      int e;
-	      int steps;
 
 	      if ((s->type < SK_MAGIC) && FD_ISSET(s->fd, &rd) && s->rx_hook)
 		{
@@ -1434,9 +1649,10 @@ test_old_bird(char *path)
   struct sockaddr_un sa;
 
   fd = socket(AF_UNIX, SOCK_STREAM, 0);
-
   if (fd < 0)
     die("Cannot create socket: %m");
+  if (strlen(path) >= sizeof(sa.sun_path))
+    die("Socket path too long");
   bzero(&sa, sizeof(sa));
   sa.sun_family = AF_UNIX;
   strcpy(sa.sun_path, path);
